import importlib
import random

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal
from pydantic.error_wrappers import ValidationError

import tests.bofire.data_models.specs.api as specs
from bofire.data_models.domain.api import Features, Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    CategoricalMolecularInput,
    CategoricalOutput,
    ContinuousDescriptorInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
    MolecularInput,
    Output,
)
from bofire.data_models.molfeatures.api import (
    Fingerprints,
    FingerprintsFragments,
    Fragments,
    MordredDescriptors,
)
from bofire.data_models.objectives.api import MinimizeObjective, Objective
from bofire.data_models.surrogates.api import ScalerEnum

objective = MinimizeObjective(w=1)

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None


@pytest.mark.parametrize(
    "spec, n",
    [
        (spec, n)
        for spec in specs.features.valids
        if (spec.cls != ContinuousOutput)
        and (spec.cls != MolecularInput)
        and (spec.cls != CategoricalOutput)
        for n in [1, 5]
    ],
)
def test_sample(spec: specs.Spec, n: int):
    feat = spec.obj()
    samples = feat.sample(n=n)
    feat.validate_candidental(samples)


@pytest.mark.parametrize(
    "input_feature, expected, expected_value",
    [
        (ContinuousInput(key="k", bounds=(1, 1)), True, [1]),
        (ContinuousInput(key="k", bounds=(1, 2)), False, None),
        (ContinuousInput(key="k", bounds=(2, 3)), False, None),
        (
            ContinuousDescriptorInput(
                key="k",
                bounds=(1, 1),
                descriptors=["a", "b"],
                values=[1, 2],
            ),
            True,
            [1],
        ),
        (
            ContinuousDescriptorInput(
                key="k",
                bounds=(1, 2),
                descriptors=["a", "b"],
                values=[1, 2],
            ),
            False,
            None,
        ),
        (
            ContinuousDescriptorInput(
                key="k",
                bounds=(2, 3),
                descriptors=["a", "b"],
                values=[1, 2],
            ),
            False,
            None,
        ),
    ],
)
def test_continuous_input_feature_is_fixed(input_feature, expected, expected_value):
    assert input_feature.is_fixed() == expected
    assert input_feature.fixed_value() == expected_value


def test_continuous_input_invalid_stepsize():
    with pytest.raises(ValueError):
        ContinuousInput(key="a", bounds=(1, 1), stepsize=0)
    with pytest.raises(ValueError):
        ContinuousInput(key="a", bounds=(0, 5), stepsize=0.3)
    with pytest.raises(ValueError):
        ContinuousInput(key="a", bounds=(0, 1), stepsize=1)


def test_continuous_input_round():
    feature = ContinuousInput(key="a", bounds=(0, 5))
    values = pd.Series([1.0, 1.3, 0.55])
    assert_series_equal(values, feature.round(values))
    feature = ContinuousInput(key="a", bounds=(0, 5), stepsize=0.25)
    assert_series_equal(pd.Series([1.0, 1.25, 0.5]), feature.round(values))
    feature = ContinuousInput(key="a", bounds=(0, 5), stepsize=0.1)
    assert_series_equal(pd.Series([1.0, 1.3, 0.5]), feature.round(values))


@pytest.mark.parametrize(
    "input_feature, expected",
    [
        (
            ContinuousInput(key="if1", bounds=(0.5, 4)),
            (0.5, 4.0),
        ),
        (ContinuousInput(key="if1", bounds=(2.5, 2.9)), (1, 3.0)),
        (ContinuousInput(key="if2", bounds=(1, 3)), (1, 3.0)),
        (ContinuousInput(key="if2", bounds=(1, 1)), (1, 1.0)),
    ],
)
def test_continuous_input_feature_get_bounds(input_feature, expected):
    experiments = pd.DataFrame({"if1": [1.0, 2.0, 3.0], "if2": [1.0, 1.0, 1.0]})
    lower, upper = input_feature.get_bounds(values=experiments[input_feature.key])
    assert (lower[0], upper[0]) == expected
    lower, upper = input_feature.get_bounds()
    assert (lower[0], upper[0]) == (
        input_feature.lower_bound,
        input_feature.upper_bound,
    )


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([random.uniform(3.0, 5.3) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([random.uniform(3.0, 5.3) for _ in range(20)]),
            False,
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([random.uniform(3.0, 5.3) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([random.uniform(3.0, 5.3) for _ in range(20)]),
            False,
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([3.0, 3.0, 3.0]),
            False,
        ),
    ],
)
def test_continuous_input_feature_validate_valid(input_feature, values, strict):
    input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([3.0, "mama"]),
            True,
        ),
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([3.0, "mama"]),
            False,
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([3.0, 3.0, 3.0]),
            True,
        ),
    ],
)
def test_continuous_input_feature_validate_invalid(input_feature, values, strict):
    with pytest.raises(ValueError):
        input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([random.uniform(3.0, 5.3) for _ in range(20)]),
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([3.0, 3.0, 3.0]),
        ),
    ],
)
def test_continuous_input_feature_validate_candidental_valid(input_feature, values):
    input_feature.validate_candidental(values)


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([3.1, "a"]),
        ),
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([2.9, 4.0]),
        ),
        (
            specs.features.valid(ContinuousInput).obj(),
            pd.Series([4.0, 6]),
        ),
        (
            specs.features.valid(ContinuousInput).obj(bounds=(3, 3)),
            pd.Series([3.1, 3.2, 3.4]),
        ),
    ],
)
def test_continuous_input_feature_validate_candidental_invalid(input_feature, values):
    with pytest.raises(ValueError):
        input_feature.validate_candidental(values)


@pytest.mark.parametrize(
    "feature, xt, expected",
    [
        (
            ContinuousInput(key="a", bounds=(0, 10)),
            pd.Series(np.linspace(0, 1, 11)),
            np.linspace(0, 10, 11),
        ),
        (
            ContinuousInput(key="a", bounds=(-10, 20)),
            pd.Series(np.linspace(0, 1)),
            np.linspace(-10, 20),
        ),
    ],
)
def test_continuous_input_feature_from_unit_range(feature, xt, expected):
    x = feature.from_unit_range(xt)
    assert np.allclose(x.values, expected)


@pytest.mark.parametrize(
    "feature, x, expected, real",
    [
        (
            ContinuousInput(key="a", bounds=(0, 10)),
            pd.Series(np.linspace(0, 10, 11)),
            np.linspace(0, 1, 11),
            True,
        ),
        (
            ContinuousInput(key="a", bounds=(-10, 20)),
            pd.Series(np.linspace(-10, 20)),
            np.linspace(0, 1),
            True,
        ),
        (
            ContinuousInput(key="a", bounds=(0, 10)),
            pd.Series(np.linspace(0, 10, 11)),
            np.linspace(0, 1, 11),
            False,
        ),
        (
            ContinuousInput(key="a", bounds=(-10, 20)),
            pd.Series(np.linspace(-10, 20)),
            np.linspace(0, 1),
            False,
        ),
        (
            ContinuousInput(key="a", bounds=(0, 9)),
            pd.Series(np.linspace(0, 10, 11)),
            np.linspace(0, 1, 11),
            True,
        ),
        (
            ContinuousInput(key="a", bounds=(0, 9)),
            pd.Series(np.linspace(0, 10, 11)),
            np.linspace(0, 10 / 9, 11),
            False,
        ),
    ],
)
def test_continuous_input_feature_to_unit_range(feature, x, expected, real):
    xt = feature.to_unit_range(x)
    assert np.allclose(xt.values, expected, real)


@pytest.mark.parametrize(
    "input_feature, expected, expected_value",
    [
        (specs.features.valid(DiscreteInput).obj(values=[1, 2, 3]), False, None),
    ],
)
def test_discrete_input_feature_is_fixed(input_feature, expected, expected_value):
    print(input_feature)
    assert input_feature.is_fixed() == expected
    assert input_feature.fixed_value() == expected_value


@pytest.mark.parametrize(
    "input_feature, expected_lower, expected_upper",
    [
        (
            specs.features.valid(DiscreteInput).obj(values=[1.0, 2.0, 3.0]),
            1,
            3,
        ),
    ],
)
def test_discrete_input_feature_bounds(input_feature, expected_lower, expected_upper):
    assert input_feature.upper_bound == expected_upper
    assert input_feature.lower_bound == expected_lower


@pytest.mark.parametrize(
    "input_feature, expected",
    [
        (
            DiscreteInput(key="if1", values=[2.0, 3.0]),
            (1.0, 4.0),
        ),
        (
            DiscreteInput(key="if1", values=[0.0, 3.0]),
            (0.0, 4.0),
        ),
        (
            DiscreteInput(key="if1", values=[2.0, 5.0]),
            (1.0, 5.0),
        ),
    ],
)
def test_discrete_input_feature_get_bounds(input_feature, expected):
    experiments = pd.DataFrame(
        {"if1": [1.0, 2.0, 3.0, 4.0], "if2": [1.0, 1.0, 1.0, 1.0]}
    )
    lower, upper = input_feature.get_bounds(values=experiments[input_feature.key])
    assert (lower[0], upper[0]) == expected
    lower, upper = input_feature.get_bounds()
    assert (lower[0], upper[0]) == (
        input_feature.lower_bound,
        input_feature.upper_bound,
    )


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(DiscreteInput).obj(values=[1, 2, 3]),
            pd.Series([random.choice([1, 2, 3]) for _ in range(20)]),
        ),
    ],
)
def test_discrete_input_feature_validate_candidental_valid(input_feature, values):
    input_feature.validate_candidental(values)


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(DiscreteInput).obj(values=[1, 2]),
            pd.Series([1, 2, 3]),
        ),
    ],
)
def test_discrete_input_feature_validate_candidental_invalid(input_feature, values):
    with pytest.raises(ValueError):
        input_feature.validate_candidental(values)


def test_from_continuous():
    d = DiscreteInput(key="d", values=[1, 2, 3])

    continuous_values = pd.DataFrame(
        columns=["d"],
        data=[1.8, 1.7, 2.9, 1.9],
    )
    samples = d.from_continuous(continuous_values)
    assert np.all(samples == pd.Series([2, 2, 3, 2]))


@pytest.mark.parametrize(
    "input_feature, expected",
    [
        (
            CategoricalInput(key="if1", categories=["a", "b"], allowed=[True, True]),
            ["a", "b"],
        ),
        (
            CategoricalInput(key="if2", categories=["a", "b"], allowed=[True, True]),
            ["a", "b"],
        ),
        (
            CategoricalInput(key="if3", categories=["a", "b"], allowed=[True, False]),
            ["a"],
        ),
        (
            CategoricalInput(key="if4", categories=["a", "b"], allowed=[False, True]),
            ["b"],
        ),
    ],
)
def test_categorical_input_feature_get_possible_categories(input_feature, expected):
    experiments = pd.DataFrame(
        {"if1": ["a", "b"], "if2": ["a", "a"], "if3": ["a", "a"], "if4": ["b", "b"]}
    )
    categories = input_feature.get_possible_categories(experiments[input_feature.key])
    assert categories == expected


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(CategoricalInput).obj(categories=["a", "b", "c"]),
            pd.Series([random.choice(["a", "b", "c"]) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(categories=["a", "b", "c"]),
            pd.Series([random.choice(["a", "b", "c"]) for _ in range(20)]),
            False,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["a", "b", "c"],
                allowed=[True, False, False],
            ),
            pd.Series([random.choice(["a", "b", "c"]) for _ in range(20)]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["1", "2", "3"],
                allowed=[True, False, False],
            ),
            pd.Series([random.choice([1, 2, 3]) for _ in range(20)]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            False,
        ),
    ],
)
def test_categorical_input_feature_validate_valid(input_feature, values, strict):
    input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(CategoricalInput).obj(categories=["a", "b", "c"]),
            pd.Series(["a", "b", "c", "d"]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(categories=["a", "b", "c"]),
            pd.Series(["a", "b", "c", "d"]),
            False,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["a", "b", "c"],
                allowed=[True, False, False],
            ),
            pd.Series(["a", "a"]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["a", "b", "c"],
                allowed=[True, False, False],
            ),
            pd.Series(["a", "b"]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["1", "2", "3"],
                allowed=[True, False, False],
            ),
            pd.Series([1, 2]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["one", "two", "three"],
                allowed=[True, False, False],
            ),
            pd.Series([1, 2, 3]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
    ],
)
def test_categorical_input_feature_validate_invalid(input_feature, values, strict):
    with pytest.raises(ValueError):
        input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[True, True, True],
            ),
            pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["a", "b"], allowed=[True, False]
            ),
            pd.Series(["a", "a"]),
        ),
    ],
)
def test_categorical_input_feature_validate_candidental_valid(input_feature, values):
    input_feature.validate_candidental(values)


@pytest.mark.parametrize(
    "input_feature, values",
    [
        (
            specs.features.valid(CategoricalInput).obj(categories=["a", "b", "c"]),
            pd.Series(["a", "b", "c", "d"]),
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["a", "b"], allowed=[True, False]
            ),
            pd.Series(["a", "b"]),
        ),
    ],
)
def test_categorical_input_feature_validate_candidental_invalid(input_feature, values):
    with pytest.raises(ValueError):
        input_feature.validate_candidental(values)


@pytest.mark.parametrize(
    "key, categories, samples",
    [
        ("c", ["B", "A", "C"], ["A", "A", "C", "B"]),
        ("c_alpha", ["B_b", "_A_a", "C_c_"], ["_A_a", "_A_a", "C_c_", "B_b"]),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
            [
                "__c_alpha_c___A_a",
                "__c_alpha_c___A_a",
                "__c_alpha_c__C_c_",
                "__c_alpha_c__B_b",
            ],
        ),
    ],
)
def test_categorical_to_one_hot_encoding(key, categories, samples):
    c = CategoricalInput(key=key, categories=categories)
    samples = pd.Series(samples)
    t_samples = c.to_onehot_encoding(samples)
    assert_frame_equal(
        t_samples,
        pd.DataFrame(
            data=[[0.0, 1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0], [1.0, 0.0, 0.0]],
            columns=[f"{key}_{cat_str}" for cat_str in categories],
        ),
    )
    untransformed = c.from_onehot_encoding(t_samples)
    assert np.all(samples == untransformed)


@pytest.mark.parametrize(
    "key, categories",
    [
        ("c", ["B", "A", "C"]),
        ("c_alpha", ["B_b", "_A_a", "C_c_"]),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
        ),
    ],
)
def test_categorical_from_one_hot_encoding(key, categories):
    c = CategoricalInput(key=key, categories=categories)
    one_hot_values = pd.DataFrame(
        columns=[f"{key}_{cat_str}" for cat_str in categories] + ["misc"],
        data=[[0.9, 0.4, 0.2, 6], [0.8, 0.7, 0.9, 9]],
    )
    samples = c.from_onehot_encoding(one_hot_values)
    assert np.all(samples == pd.Series([categories[0], categories[2]]))


def test_categorical_from_one_hot_encoding_invalid():
    c = CategoricalInput(key="c", categories=["B", "A", "C"])
    one_hot_values = pd.DataFrame(
        columns=["c_B", "c_A", "misc"],
        data=[
            [
                0.9,
                0.4,
                0.2,
            ],
            [0.8, 0.7, 0.9],
        ],
    )
    with pytest.raises(ValueError):
        c.from_onehot_encoding(one_hot_values)


@pytest.mark.parametrize(
    "key, categories, samples",
    [
        ("c", ["B", "A", "C"], ["A", "A", "C", "B"]),
        ("c_alpha", ["B_b", "_A_a", "C_c_"], ["_A_a", "_A_a", "C_c_", "B_b"]),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
            [
                "__c_alpha_c___A_a",
                "__c_alpha_c___A_a",
                "__c_alpha_c__C_c_",
                "__c_alpha_c__B_b",
            ],
        ),
    ],
)
def test_categorical_to_dummy_encoding(key, categories, samples):
    c = CategoricalInput(key=key, categories=categories)
    samples = pd.Series(samples)
    t_samples = c.to_dummy_encoding(samples)
    assert_frame_equal(
        t_samples,
        pd.DataFrame(
            data=[[1.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.0, 0.0]],
            columns=[f"{key}_{cat_str}" for cat_str in categories[1:]],
        ),
    )
    untransformed = c.from_dummy_encoding(t_samples)
    assert np.all(samples == untransformed)


@pytest.mark.parametrize(
    "key, categories",
    [
        ("c", ["B", "A", "C"]),
        ("c_alpha", ["B_b", "_A_a", "C_c_"]),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
        ),
    ],
)
def test_categorical_from_dummy_encoding(key, categories):
    c = CategoricalInput(key=key, categories=categories)
    one_hot_values = pd.DataFrame(
        columns=[f"{key}_{cat_str}" for cat_str in categories[1:]] + ["misc"],
        data=[[0.9, 0.05, 6], [0.1, 0.1, 9]],
    )
    samples = c.from_dummy_encoding(one_hot_values)
    assert np.all(samples == pd.Series([categories[1], categories[0]]))


def test_categorical_to_label_encoding():
    c = CategoricalInput(key="c", categories=["B", "A", "C"])
    samples = pd.Series(["A", "A", "C", "B"])
    t_samples = c.to_ordinal_encoding(samples)
    assert_series_equal(t_samples, pd.Series([1, 1, 2, 0], name="c"))
    untransformed = c.from_ordinal_encoding(t_samples)
    assert np.all(samples == untransformed)


@pytest.mark.parametrize(
    "feature, transform_type, values, expected",
    [
        (
            CategoricalInput(key="c", categories=["B", "A", "C"]),
            CategoricalEncodingEnum.ORDINAL,
            None,
            (0, 2),
        ),
        (
            CategoricalInput(key="c", categories=["B", "A", "C"]),
            CategoricalEncodingEnum.ONE_HOT,
            None,
            ([0, 0, 0], [1, 1, 1]),
        ),
        (
            CategoricalInput(
                key="c", categories=["B", "A", "C"], allowed=[True, False, True]
            ),
            CategoricalEncodingEnum.ONE_HOT,
            pd.Series(["A", "B", "C"]),
            ([0, 0, 0], [1, 1, 1]),
        ),
        (
            CategoricalInput(
                key="c", categories=["B", "A", "C"], allowed=[True, False, True]
            ),
            CategoricalEncodingEnum.ONE_HOT,
            None,
            ([0, 0, 0], [1, 0, 1]),
        ),
        (
            CategoricalInput(key="c", categories=["B", "A", "C"]),
            CategoricalEncodingEnum.DUMMY,
            None,
            ([0, 0], [1, 1]),
        ),
    ],
)
def test_categorical_get_bounds(feature, transform_type, values, expected):
    lower, upper = feature.get_bounds(transform_type=transform_type, values=values)
    assert np.allclose(lower, expected[0])
    assert np.allclose(upper, expected[1])
    # test the same for the categorical with descriptor
    f = CategoricalDescriptorInput(
        key="c",
        categories=feature.categories,
        allowed=feature.allowed,
        descriptors=["alpha", "beta"],
        values=[[1, 2], [3, 4], [5, 6]],
    )
    lower, upper = f.get_bounds(transform_type=transform_type, values=values)
    assert np.allclose(lower, expected[0])
    assert np.allclose(upper, expected[1])


@pytest.mark.parametrize(
    "key, categories, samples_in, descriptors",
    [
        ("c", ["B", "A", "C"], ["A", "A", "C", "B"], ["d1", "d2"]),
        (
            "c_alpha",
            ["B_b", "_A_a", "C_c_"],
            ["_A_a", "_A_a", "C_c_", "B_b"],
            ["_d1_d", "d2_d_2_"],
        ),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
            [
                "__c_alpha_c___A_a",
                "__c_alpha_c___A_a",
                "__c_alpha_c__C_c_",
                "__c_alpha_c__B_b",
            ],
            ["__c_alpha_c__d1_d", "__c_alpha_c_d2_d_2_"],
        ),
    ],
)
def test_categorical_descriptor_to_descriptor_encoding(
    key, categories, samples_in, descriptors
):
    c = CategoricalDescriptorInput(
        key=key,
        categories=categories,
        descriptors=descriptors,
        values=[[1, 2], [3, 4], [5, 6]],
    )
    samples = pd.Series(samples_in)
    t_samples = c.to_descriptor_encoding(samples)
    assert_frame_equal(
        t_samples,
        pd.DataFrame(
            data=[[3.0, 4.0], [3.0, 4.0], [5.0, 6.0], [1.0, 2.0]],
            columns=[f"{key}_{des_str}" for des_str in descriptors],
        ),
    )
    untransformed = c.from_descriptor_encoding(t_samples)
    assert np.all(samples == untransformed)


@pytest.mark.parametrize(
    "key, categories, descriptors",
    [
        ("c", ["B", "A", "C"], ["d1", "d2"]),
        ("c_alpha", ["B_b", "_A_a", "C_c_"], ["_d1_d", "d2_d_2_"]),
        (
            "__c_alpha_c_",
            ["__c_alpha_c__B_b", "__c_alpha_c___A_a", "__c_alpha_c__C_c_"],
            ["__c_alpha_c__d1_d", "__c_alpha_c_d2_d_2_"],
        ),
    ],
)
def test_categorical_descriptor_from_descriptor_encoding(key, categories, descriptors):
    c1 = CategoricalDescriptorInput(
        key=key,
        categories=categories,
        descriptors=descriptors,
        values=[[1, 2], [3, 4], [5, 6]],
    )
    descriptor_values = pd.DataFrame(
        columns=[f"{key}_{des_str}" for des_str in descriptors] + ["misc"],
        data=[[1.05, 2.5, 6], [4, 4.5, 9]],
    )
    samples = c1.from_descriptor_encoding(descriptor_values)
    print(samples)
    assert np.all(samples == pd.Series([categories[0], categories[1]]))

    c2 = CategoricalDescriptorInput(
        key=key,
        categories=categories,
        descriptors=descriptors,
        values=[[1, 2], [3, 4], [5, 6]],
        allowed=[False, True, True],
    )

    samples = c2.from_descriptor_encoding(descriptor_values)
    print(samples)
    assert np.all(samples == pd.Series([categories[1], categories[1]]))


def test_categorical_descriptor_to_descriptor_encoding_1d():
    c = CategoricalDescriptorInput(
        key="c",
        categories=["B", "A", "C"],
        descriptors=["d1"],
        values=[[1], [3], [5]],
    )
    samples = pd.Series(["A", "A", "C", "B"])
    t_samples = c.to_descriptor_encoding(samples)
    assert_frame_equal(
        t_samples,
        pd.DataFrame(
            data=[[3.0], [3.0], [5.0], [1.0]],
            columns=["c_d1"],
        ),
    )
    untransformed = c.from_descriptor_encoding(t_samples)
    assert np.all(samples == untransformed)


@pytest.mark.parametrize(
    "input_feature, expected_with_values, expected",
    [
        (
            CategoricalDescriptorInput(
                key="if1",
                categories=["a", "b"],
                allowed=[True, True],
                descriptors=["alpha", "beta"],
                values=[[1, 2], [3, 4]],
            ),
            ([1, 2], [3, 4]),
            ([1, 2], [3, 4]),
        ),
        (
            CategoricalDescriptorInput(
                key="if2",
                categories=["a", "b", "c"],
                allowed=[True, False, True],
                descriptors=["alpha", "beta"],
                values=[[1, 2], [3, 4], [1, 5]],
            ),
            ([1, 2], [3, 5]),
            ([1, 2], [1, 5]),
        ),
        # (CategoricalInput(key="if2", categories = ["a","b"], allowed = [True, True]), ["a","b"]),
        # (CategoricalInput(key="if3", categories = ["a","b"], allowed = [True, False]), ["a"]),
        # (CategoricalInput(key="if4", categories = ["a","b"], allowed = [True, False]), ["a", "b"]),
        # (ContinuousInput(key="if1", lower_bound=2.5, upper_bound=2.9), (1,3.)),
        # (ContinuousInput(key="if2", lower_bound=1., upper_bound=3.), (1,3.)),
        # (ContinuousInput(key="if2", lower_bound=1., upper_bound=1.), (1,1.)),
    ],
)
def test_categorical_descriptor_feature_get_bounds(
    input_feature, expected_with_values, expected
):
    experiments = pd.DataFrame(
        {"if1": ["a", "b"], "if2": ["a", "c"], "if3": ["a", "a"], "if4": ["b", "b"]}
    )
    lower, upper = input_feature.get_bounds(
        transform_type=CategoricalEncodingEnum.DESCRIPTOR,
        values=experiments[input_feature.key],
    )
    assert np.allclose(lower, expected_with_values[0])
    assert np.allclose(upper, expected_with_values[1])
    lower, upper = input_feature.get_bounds(
        transform_type=CategoricalEncodingEnum.DESCRIPTOR,
        values=None,
    )
    assert np.allclose(lower, expected[0])
    assert np.allclose(upper, expected[1])


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"]
            ),
            pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"]
            ),
            pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            False,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[True, False, False],
            ),
            pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[True, False, False],
            ),
            pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            False,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[True, False, False],
            ),
            pd.Series(["c1", "c1"]),
            False,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[False, True, True],
                descriptors=["d1", "d2"],
                values=[
                    [1, 2],
                    [3, 7],
                    [3, 1],
                ],
            ),
            pd.Series(["c2", "c3"]),
            False,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["1", "2", "3"],
                allowed=[True, False, False],
            ),
            pd.Series([random.choice([1, 2, 3]) for _ in range(20)]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            False,
        ),
    ],
)
def test_categorical_descriptor_input_feature_validate_valid(
    input_feature, values, strict
):
    input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, values, strict",
    [
        (
            specs.features.valid(CategoricalDescriptorInput).obj(),
            pd.Series(["c1", "c4"]),
            True,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(),
            pd.Series(["c1", "c4"]),
            False,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[True, False, False],
                descriptors=["d1", "d2"],
                values=[
                    [1, 2],
                    [3, 7],
                    [5, 1],
                ],
            ),
            pd.Series(["c1", "c1"]),
            True,
        ),
        (
            specs.features.valid(CategoricalDescriptorInput).obj(
                categories=["c1", "c2", "c3"],
                allowed=[False, True, True],
                descriptors=["d1", "d2"],
                values=[
                    [1, 2],
                    [3, 7],
                    [3, 1],
                ],
            ),
            pd.Series(["c2", "c3"]),
            True,
        ),
        (
            specs.features.valid(CategoricalInput).obj(
                categories=["1", "2", "3"],
                allowed=[True, False, False],
            ),
            pd.Series([1, 2]),
            # CategoricalInput(**VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC),
            # pd.Series([random.choice(["c1", "c2", "c3"]) for _ in range(20)]),
            True,
        ),
    ],
)
def test_categorical_descriptor_input_feature_validate_invalid(
    input_feature, values, strict
):
    with pytest.raises(ValueError):
        input_feature.validate_experimental(values, strict)


@pytest.mark.parametrize(
    "input_feature, expected, expected_value, transform_type",
    [
        (
            CategoricalInput(key="k", categories=categories, allowed=allowed),
            expected,
            expected_value,
            transform_type,
        )
        for categories, allowed, expected, expected_value, transform_type in [
            (["1", "2"], None, False, None, None),
            (["1", "2", "3", "4"], [True, False, False, False], True, ["1"], None),
            (["1", "2", "3", "4"], [True, True, False, True], False, None, None),
            (
                ["1", "2", "3", "4"],
                [True, False, False, False],
                True,
                [0],
                CategoricalEncodingEnum.ORDINAL,
            ),
            (
                ["1", "2", "3", "4"],
                [True, False, False, False],
                True,
                [1, 0, 0, 0],
                CategoricalEncodingEnum.ONE_HOT,
            ),
            (
                ["1", "2", "3", "4"],
                [True, False, False, False],
                True,
                [0, 0, 0],
                CategoricalEncodingEnum.DUMMY,
            ),
        ]
    ]
    + [
        (
            CategoricalDescriptorInput(
                key="k",
                categories=["1", "2", "3"],
                allowed=[True, False, False],
                descriptors=["alpha", "beta"],
                values=[[1, 2], [3, 4], [5, 6]],
            ),
            expected,
            expected_value,
            transform_type,
        )
        for expected, expected_value, transform_type in [
            (True, [1, 2], CategoricalEncodingEnum.DESCRIPTOR)
        ]
    ],
)
def test_categorical_input_feature_is_fixed(
    input_feature, expected, expected_value, transform_type
):
    assert input_feature.is_fixed() == expected
    assert input_feature.fixed_value(transform_type) == expected_value


@pytest.mark.parametrize(
    "input_feature, expected",
    [
        (
            CategoricalInput(key="k", categories=categories, allowed=allowed),
            expected,
        )
        for categories, allowed, expected in [
            (["a", "b", "c"], [True, True, True], ["a", "b", "c"]),
            (["a", "b", "c"], [False, True, True], ["b", "c"]),
        ]
    ],
)
def test_categorical_input_feature_allowed_categories(input_feature, expected):
    assert input_feature.get_allowed_categories() == expected


@pytest.mark.parametrize(
    "input_feature, expected",
    [
        (
            CategoricalInput(key="k", categories=categories, allowed=allowed),
            expected,
        )
        for categories, allowed, expected in [
            (["a", "b", "c"], [True, True, True], []),
            (["a", "b", "c"], [False, True, True], ["a"]),
        ]
    ],
)
def test_categorical_input_feature_forbidden_categories(input_feature, expected):
    assert input_feature.get_forbidden_categories() == expected


@pytest.mark.parametrize(
    "categories, descriptors, values",
    [
        (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3], [4, 5, 6]]),
        (
            ["c1", "c2", "c3", "c4"],
            ["d1", "d2", "d3"],
            [
                [1, 2, 3],
                [4, 5, 6],
                [4, 5, 6],
                [4, 5, 6],
            ],
        ),
    ],
)
def test_categorical_descriptor_input_feature_as_dataframe(
    categories, descriptors, values
):
    f = CategoricalDescriptorInput(
        key="k", categories=categories, descriptors=descriptors, values=values
    )
    df = f.to_df()
    assert len(df.columns) == len(descriptors)
    assert len(df) == len(categories)
    assert df.values.tolist() == values


@pytest.mark.parametrize(
    "descriptors, values",
    [
        (["a", "b"], [1.0, 2.0]),
        (["a", "b", "c"], [1.0, 2.0, 3.0]),
    ],
)
def test_continuous_descriptor_input_feature_as_dataframe(descriptors, values):
    f = ContinuousDescriptorInput(
        key="k",
        bounds=(1, 2),
        descriptors=descriptors,
        values=values,
    )
    df = f.to_df()
    assert len(df.columns) == len(descriptors)
    assert len(df) == 1
    assert df.values.tolist()[0] == values


@pytest.mark.parametrize(
    "categories, descriptors, values",
    [
        (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3], [4, 5, 6]]),
        (
            ["c1", "c2", "c3", "c4"],
            ["d1", "d2", "d3"],
            [
                [1, 2, 3],
                [4, 5, 6],
                [4, 5, 6],
                [4, 5, 6],
            ],
        ),
    ],
)
def test_categorical_descriptor_input_feature_from_dataframe(
    categories, descriptors, values
):
    df = pd.DataFrame.from_dict(
        dict(zip(categories, values)),
        orient="index",
        columns=descriptors,
    )
    f = CategoricalDescriptorInput.from_df("k", df)
    assert f.categories == categories
    assert f.descriptors == descriptors
    assert f.values == values


cont = specs.features.valid(ContinuousInput).obj()
cat = specs.features.valid(CategoricalInput).obj()
cat_ = specs.features.valid(CategoricalDescriptorInput).obj()
out = specs.features.valid(ContinuousOutput).obj()


@pytest.mark.parametrize(
    "unsorted_list, sorted_list",
    [
        (
            [cont, cat_, cat, out],
            [cont, cat_, cat, out],
        ),
        (
            [cont, cat_, cat, out, cat_, out],
            [cont, cat_, cat_, cat, out, out],
        ),
        (
            [cont, out],
            [cont, out],
        ),
        (
            [out, cont],
            [cont, out],
        ),
    ],
)
def test_feature_sorting(unsorted_list, sorted_list):
    assert sorted(unsorted_list) == sorted_list


# test features container
if1 = specs.features.valid(ContinuousInput).obj(key="if1")
if2 = specs.features.valid(ContinuousInput).obj(key="if2")
if3 = specs.features.valid(ContinuousInput).obj(key="if3", bounds=(3, 3))
if4 = specs.features.valid(CategoricalInput).obj(
    key="if4", categories=["a", "b"], allowed=[True, False]
)
if5 = specs.features.valid(DiscreteInput).obj(key="if5")
if7 = specs.features.valid(CategoricalInput).obj(
    key="if7",
    categories=["c", "d", "e"],
    allowed=[True, False, False],
)


of1 = specs.features.valid(ContinuousOutput).obj(key="of1")
of2 = specs.features.valid(ContinuousOutput).obj(key="of2")
of3 = specs.features.valid(ContinuousOutput).obj(key="of3", objective=None)

inputs = Inputs(features=[if1, if2])
outputs = Outputs(features=[of1, of2])
features = Features(features=[if1, if2, of1, of2])


@pytest.mark.parametrize(
    "FeatureContainer, features",
    [
        (Features, ["s"]),
        (Features, [specs.features.valid(ContinuousInput).obj(), 5]),
        (Inputs, ["s"]),
        (Inputs, [specs.features.valid(ContinuousInput).obj(), 5]),
        (
            Inputs,
            [
                specs.features.valid(ContinuousInput).obj(),
                specs.features.valid(ContinuousOutput).obj(),
            ],
        ),
        (Outputs, ["s"]),
        (Outputs, [specs.features.valid(ContinuousOutput).obj(), 5]),
        (
            Outputs,
            [
                specs.features.valid(ContinuousOutput).obj(),
                specs.features.valid(ContinuousInput).obj(),
            ],
        ),
    ],
)
def test_features_invalid_feature(FeatureContainer, features):
    with pytest.raises((ValueError, TypeError, KeyError, ValidationError)):
        FeatureContainer(features=features)


@pytest.mark.parametrize(
    "features1, features2, expected_type",
    [
        [inputs, inputs, Inputs],
        [outputs, outputs, Outputs],
        [inputs, outputs, Features],
        [outputs, inputs, Features],
        [features, outputs, Features],
        [features, inputs, Features],
        [outputs, features, Features],
        [inputs, features, Features],
    ],
)
def test_features_plus(features1, features2, expected_type):
    returned = features1 + features2
    assert type(returned) == expected_type
    assert len(returned) == (len(features1) + len(features2))


@pytest.mark.parametrize(
    "features, FeatureType, exact, expected",
    [
        (features, Feature, False, [if1, if2, of1, of2]),
        (features, Output, False, [of1, of2]),
        (inputs, ContinuousInput, False, [if1, if2]),
        (outputs, ContinuousOutput, False, [of1, of2]),
    ],
)
def test_constraints_get(features, FeatureType, exact, expected):
    returned = features.get(FeatureType, exact=exact)
    assert returned.features == expected
    for i in range(len(expected)):
        assert id(expected[i]) == id(returned[i])
    assert type(returned) == type(features)


@pytest.mark.parametrize(
    "features, FeatureType, exact, expected",
    [
        (features, Feature, False, ["if1", "if2", "of1", "of2"]),
        (features, Output, False, ["of1", "of2"]),
        (inputs, ContinuousInput, False, ["if1", "if2"]),
        (outputs, ContinuousOutput, False, ["of1", "of2"]),
    ],
)
def test_features_get_keys(features, FeatureType, exact, expected):
    assert features.get_keys(FeatureType, exact=exact) == expected


@pytest.mark.parametrize(
    "features, key, expected",
    [
        (features, "if1", if1),
        (outputs, "of1", of1),
        (inputs, "if1", if1),
    ],
)
def test_features_get_by_key(features, key, expected):
    returned = features.get_by_key(key)
    assert returned.key == expected.key
    assert id(returned) == id(expected)


def test_features_get_by_keys():
    keys = ["of2", "if1"]
    feats = features.get_by_keys(keys)
    assert feats[0].key == "if1"
    assert feats[1].key == "of2"


@pytest.mark.parametrize(
    "features, key",
    [
        (features, "if133"),
        (outputs, "of3331"),
        (inputs, "if1333333"),
    ],
)
def test_features_get_by_key_invalid(features, key):
    with pytest.raises(KeyError):
        features.get_by_key(key)


@pytest.mark.parametrize(
    "features, expected",
    [
        (Inputs(features=[if1, if2]), []),
        (Inputs(features=[if1, if2, if3, if4, if5]), [if3, if4]),
    ],
)
def test_inputs_get_fixed(features, expected):
    returned = features.get_fixed()
    assert isinstance(features, Inputs)
    assert returned.features == expected
    for i in range(len(expected)):
        assert id(expected[i]) == id(returned[i])


@pytest.mark.parametrize(
    "features, expected",
    [
        (Inputs(features=[if1, if2]), [if1, if2]),
        (Inputs(features=[if1, if2, if3, if4, if5]), [if1, if2, if5]),
    ],
)
def test_inputs_get_free(features, expected):
    returned = features.get_free()
    assert isinstance(features, Inputs)
    assert returned.features == expected
    for i in range(len(expected)):
        assert id(expected[i]) == id(returned[i])


@pytest.mark.parametrize(
    "features, num_samples, method",
    [
        (features, num_samples, method)
        for features in [
            inputs,
            Inputs(features=[if1, if2, if3, if4, if5, if7]),
        ]
        for num_samples in [1, 2, 1024]
        for method in ["UNIFORM", "SOBOL", "LHS"]
    ],
)
def test_inputs_sample(features: Inputs, num_samples, method):
    samples = features.sample(num_samples, method=method)
    assert samples.shape == (num_samples, len(features))
    assert list(samples.columns) == features.get_keys()


@pytest.mark.parametrize(
    "specs",
    [
        ({"x4": CategoricalEncodingEnum.ONE_HOT}),
        ({"x1": CategoricalEncodingEnum.ONE_HOT}),
        ({"x2": ScalerEnum.NORMALIZE}),
        ({"x2": CategoricalEncodingEnum.DESCRIPTOR}),
        ({"x1": Fingerprints()}),
        ({"x2": Fragments()}),
        ({"x3": FingerprintsFragments()}),
        ({"x3": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"])}),
    ],
)
def test_inputs_validate_transform_specs_invalid(specs):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4]],
            ),
        ]
    )
    with pytest.raises(ValueError):
        inps._validate_transform_specs(specs)


@pytest.mark.parametrize(
    "specs",
    [
        ({"x2": CategoricalEncodingEnum.ONE_HOT}),
        ({"x3": CategoricalEncodingEnum.ONE_HOT}),
        ({"x3": CategoricalEncodingEnum.DESCRIPTOR}),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
            }
        ),
    ],
)
def test_inputs_validate_transform_valid(specs):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4]],
            ),
        ]
    )
    inps._validate_transform_specs(specs)


@pytest.mark.parametrize(
    "specs",
    [
        # ({"x2": CategoricalEncodingEnum.ONE_HOT}),
        # ({"x3": CategoricalEncodingEnum.DESCRIPTOR}),
        ({"x4": CategoricalEncodingEnum.ONE_HOT}),
        ({"x4": ScalerEnum.NORMALIZE}),
        ({"x4": CategoricalEncodingEnum.DESCRIPTOR}),
        # (
        #    {
        #        "x2": CategoricalEncodingEnum.ONE_HOT,
        #        "x3": CategoricalEncodingEnum.DESCRIPTOR,
        #    }
        # ),
    ],
)
def test_inputs_validate_transform_specs_molecular_input_invalid(specs):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4]],
            ),
            MolecularInput(key="x4"),
        ]
    )
    with pytest.raises(ValueError):
        inps._validate_transform_specs(specs)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "specs",
    [
        ({"x4": Fingerprints()}),
        ({"x4": Fragments()}),
        ({"x4": FingerprintsFragments()}),
        ({"x4": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"])}),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x4": Fingerprints(),
            }
        ),
        (
            {
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
                "x4": Fingerprints(),
            }
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
                "x4": Fingerprints(),
            }
        ),
    ],
)
def test_inputs_validate_transform_specs_molecular_input_valid(specs):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4]],
            ),
            MolecularInput(key="x4"),
        ]
    )
    inps._validate_transform_specs(specs)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "specs, expected_features2idx, expected_features2names",
    [
        (
            {"x2": CategoricalEncodingEnum.ONE_HOT, "x4": Fingerprints(n_bits=2048)},
            {
                "x1": (2048,),
                "x2": (2050, 2051, 2052),
                "x3": (2049,),
                "x4": tuple(range(2048)),
            },
            {
                "x1": ("x1",),
                "x2": ("x2_apple", "x2_banana", "x2_orange"),
                "x3": ("x3",),
                "x4": tuple(f"x4_fingerprint_{i}" for i in range(2048)),
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.DUMMY,
                "x4": Fragments(fragments=["fr_unbrch_alkane", "fr_thiocyan"]),
            },
            {"x1": (2,), "x2": (4, 5), "x3": (3,), "x4": (0, 1)},
            {
                "x1": ("x1",),
                "x2": ("x2_banana", "x2_orange"),
                "x3": ("x3",),
                "x4": ("x4_fr_unbrch_alkane", "x4_fr_thiocyan"),
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ORDINAL,
                "x4": FingerprintsFragments(
                    n_bits=2048, fragments=["fr_unbrch_alkane", "fr_thiocyan"]
                ),
            },
            {
                "x1": (2050,),
                "x2": (2052,),
                "x3": (2051,),
                "x4": tuple(range(2048 + 2)),
            },
            {
                "x1": ("x1",),
                "x2": ("x2",),
                "x3": ("x3",),
                "x4": tuple(
                    [f"x4_fingerprint_{i}" for i in range(2048)]
                    + ["x4_fr_unbrch_alkane", "x4_fr_thiocyan"]
                ),
            },
        ),
        (
            {
                "x3": CategoricalEncodingEnum.ONE_HOT,
                "x4": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
            },
            {"x1": (2,), "x2": (7,), "x3": (3, 4, 5, 6), "x4": (0, 1)},
            {
                "x1": ("x1",),
                "x2": ("x2",),
                "x3": ("x3_apple", "x3_banana", "x3_orange", "x3_cherry"),
                "x4": ("x4_NssCH2", "x4_ATSC2d"),
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
                "x4": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
            },
            {"x1": (2,), "x2": (5, 6, 7), "x3": (3, 4), "x4": (0, 1)},
            {
                "x1": ("x1",),
                "x2": ("x2_apple", "x2_banana", "x2_orange"),
                "x3": (
                    "x3_d1",
                    "x3_d2",
                ),
                "x4": ("x4_NssCH2", "x4_ATSC2d"),
            },
        ),
    ],
)
def test_inputs_get_transform_info(
    specs, expected_features2idx, expected_features2names
):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana", "orange"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana", "orange", "cherry"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4], [5, 6], [7, 8]],
            ),
            MolecularInput(key="x4"),
        ]
    )
    features2idx, features2names = inps._get_transform_info(specs)
    assert features2idx == expected_features2idx
    assert features2names == expected_features2names


@pytest.mark.parametrize(
    "specs",
    [
        ({"x2": CategoricalEncodingEnum.ONE_HOT}),
        ({"x2": CategoricalEncodingEnum.DUMMY}),
        ({"x2": CategoricalEncodingEnum.ORDINAL}),
        ({"x3": CategoricalEncodingEnum.ONE_HOT}),
        ({"x3": CategoricalEncodingEnum.DESCRIPTOR}),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
            }
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.ONE_HOT,
            }
        ),
        (
            {
                "x2": CategoricalEncodingEnum.DUMMY,
                "x3": CategoricalEncodingEnum.ONE_HOT,
            }
        ),
    ],
)
def test_inputs_transform(specs):
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana", "orange"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana", "orange", "cherry"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4], [5, 6], [7, 8]],
            ),
        ]
    )
    samples = inps.sample(n=100)
    samples = samples.sample(40)
    transformed = inps.transform(experiments=samples, specs=specs)
    untransformed = inps.inverse_transform(experiments=transformed, specs=specs)
    assert_frame_equal(samples, untransformed)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_input_reverse_transform_molecular():
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana", "orange"]),
            CategoricalMolecularInput(
                key="x3",
                categories=[
                    "CC(=O)Oc1ccccc1C(=O)O",
                    "c1ccccc1",
                    "[CH3][CH2][OH]",
                    "N[C@](C)(F)C(=O)O",
                ],
            ),
        ],
    )
    specs = {
        "x3": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
        "x2": CategoricalEncodingEnum.ONE_HOT,
    }
    samples = inps.sample(n=20)
    transformed = inps.transform(experiments=samples, specs=specs)
    untransformed = inps.inverse_transform(experiments=transformed, specs=specs)
    assert_frame_equal(samples, untransformed)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "specs, expected",
    [
        (
            {"x2": CategoricalEncodingEnum.ONE_HOT, "x4": Fingerprints(n_bits=32)},
            {
                "x4_fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "x4_fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "x4_fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x1": {0: 0.1, 1: 0.3, 2: 0.5, 3: 1.0},
                "x3": {0: "banana", 1: "orange", 2: "apple", 3: "cherry"},
                "x2_apple": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x2_banana": {0: 0.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x2_orange": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.DUMMY,
                "x4": Fragments(fragments=["fr_unbrch_alkane", "fr_thiocyan"]),
            },
            {
                "x4_fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x1": {0: 0.1, 1: 0.3, 2: 0.5, 3: 1.0},
                "x3": {0: "banana", 1: "orange", 2: "apple", 3: "cherry"},
                "x2_banana": {0: 0.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x2_orange": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ORDINAL,
                "x4": FingerprintsFragments(
                    n_bits=32, fragments=["fr_unbrch_alkane", "fr_thiocyan"]
                ),
            },
            {
                "x4_fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "x4_fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "x4_fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "x4_fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x4_fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x4_fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "x1": {0: 0.1, 1: 0.3, 2: 0.5, 3: 1.0},
                "x3": {0: "banana", 1: "orange", 2: "apple", 3: "cherry"},
                "x2": {0: 0, 1: 1, 2: 0, 3: 2},
            },
        ),
        (
            {
                "x2": CategoricalEncodingEnum.ONE_HOT,
                "x3": CategoricalEncodingEnum.DESCRIPTOR,
                "x4": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
            },
            {
                "x4_NssCH2": {
                    0: 0.5963718820861676,
                    1: -1.5,
                    2: -0.28395061728395066,
                    3: -8.34319526627219,
                },
                "x4_ATSC2d": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x1": {0: 0.1, 1: 0.3, 2: 0.5, 3: 1.0},
                "x3_d1": {0: 3.0, 1: 5.0, 2: 1.0, 3: 7.0},
                "x3_d2": {0: 4.0, 1: 6.0, 2: 2.0, 3: 8.0},
                "x2_apple": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "x2_banana": {0: 0.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "x2_orange": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
            },
        ),
    ],
)
def test_inputs_transform_molecular(specs, expected):
    experiments = [
        [0.1, "apple", "banana", "CC(=O)Oc1ccccc1C(=O)O", 88.0],
        [0.3, "banana", "orange", "c1ccccc1", 35.0],
        [0.5, "apple", "apple", "[CH3][CH2][OH]", 69.0],
        [1.0, "orange", "cherry", "N[C@](C)(F)C(=O)O", 20.0],
    ]
    experiments = pd.DataFrame(experiments, columns=["x1", "x2", "x3", "x4", "y"])
    experiments["valid_y"] = 1
    inps = Inputs(
        features=[
            ContinuousInput(key="x1", bounds=(0, 1)),
            CategoricalInput(key="x2", categories=["apple", "banana", "orange"]),
            CategoricalDescriptorInput(
                key="x3",
                categories=["apple", "banana", "orange", "cherry"],
                descriptors=["d1", "d2"],
                values=[[1, 2], [3, 4], [5, 6], [7, 8]],
            ),
            MolecularInput(key="x4"),
        ]
    )
    transformed = inps.transform(experiments=experiments, specs=specs)
    assert_frame_equal(transformed, pd.DataFrame.from_dict(expected))


if1 = specs.features.valid(ContinuousInput).obj(key="if1")
if2 = specs.features.valid(ContinuousInput).obj(key="if2", bounds=(3, 3))
if3 = specs.features.valid(CategoricalInput).obj(
    key="if3",
    categories=["c1", "c2", "c3"],
    allowed=[True, True, True],
)
if4 = specs.features.valid(CategoricalInput).obj(
    key="if4",
    categories=["c1", "c2", "c3"],
    allowed=[True, False, False],
)
if5 = specs.features.valid(CategoricalDescriptorInput).obj(
    key="if5",
    categories=["c1", "c2", "c3"],
    allowed=[True, False, False],
    descriptors=["d1", "d2"],
    values=[
        [1, 2],
        [3, 7],
        [5, 1],
    ],
)
if6 = specs.features.valid(CategoricalDescriptorInput).obj(
    key="if6",
    categories=["c1", "c2", "c3"],
    allowed=[True, False, False],
    descriptors=["d1", "d2"],
    values=[
        [1, 2],
        [3, 7],
        [5, 1],
    ],
)

of1 = specs.features.valid(ContinuousOutput).obj(key="of1")

inputs1 = Inputs(features=[if1, if3, if5])

inputs2 = Inputs(
    features=[
        if1,
        if2,
        if3,
        if4,
        if5,
        if6,
    ]
)


@pytest.mark.parametrize(
    "inputs, specs, expected_bounds",
    [
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.ONE_HOT,
                "if5": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [[3, 1, 2, 0, 0, 0], [5.3, 1, 2, 1, 1, 1]],
        ),
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.DUMMY,
                "if5": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [[3, 1, 2, 0, 0], [5.3, 1, 2, 1, 1]],
        ),
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.DUMMY,
                "if5": CategoricalEncodingEnum.DUMMY,
            },
            [[3, 0, 0, 0, 0], [5.3, 1, 1, 1, 1]],
        ),
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.ONE_HOT,
                "if5": CategoricalEncodingEnum.ONE_HOT,
            },
            [[3, 0, 0, 0, 0, 0, 0], [5.3, 1, 0, 0, 1, 1, 1]],
        ),
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.ORDINAL,
                "if5": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [[3, 1, 2, 0], [5.3, 1, 2, 2]],
        ),
        (
            inputs1,
            {
                "if3": CategoricalEncodingEnum.ORDINAL,
                "if5": CategoricalEncodingEnum.ORDINAL,
            },
            [[3, 0, 0], [5.3, 2, 2]],
        ),
        # new domain
        (
            inputs2,
            {
                "if3": CategoricalEncodingEnum.ONE_HOT,
                "if4": CategoricalEncodingEnum.ONE_HOT,
                "if5": CategoricalEncodingEnum.DESCRIPTOR,
                "if6": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [
                [3, 3, 1, 2, 1, 2, 0, 0, 0, 0, 0, 0],
                [
                    5.3,
                    3,
                    1,
                    2,
                    1,
                    2,
                    1,
                    1,
                    1,
                    1,
                    0,
                    0,
                ],
            ],
        ),
        (
            inputs2,
            {
                "if3": CategoricalEncodingEnum.ONE_HOT,
                "if4": CategoricalEncodingEnum.ONE_HOT,
                "if5": CategoricalEncodingEnum.ONE_HOT,
                "if6": CategoricalEncodingEnum.ONE_HOT,
            },
            [
                [3, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [5.3, 3, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0],
            ],
        ),
        (
            inputs2,
            {
                "if3": CategoricalEncodingEnum.ORDINAL,
                "if4": CategoricalEncodingEnum.ORDINAL,
                "if5": CategoricalEncodingEnum.DESCRIPTOR,
                "if6": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [
                [
                    3,
                    3,
                    1,
                    2,
                    1,
                    2,
                    0,
                    0,
                ],
                [5.3, 3, 1, 2, 1, 2, 2, 2],
            ],
        ),
        (
            inputs2,
            {
                "if3": CategoricalEncodingEnum.ORDINAL,
                "if4": CategoricalEncodingEnum.ORDINAL,
                "if5": CategoricalEncodingEnum.ORDINAL,
                "if6": CategoricalEncodingEnum.ORDINAL,
            },
            [[3, 3, 0, 0, 0, 0], [5.3, 3, 2, 2, 2, 2]],
        ),
        (
            inputs2,
            {
                "if3": CategoricalEncodingEnum.ORDINAL,
                "if4": CategoricalEncodingEnum.ONE_HOT,
                "if5": CategoricalEncodingEnum.ORDINAL,
                "if6": CategoricalEncodingEnum.DESCRIPTOR,
            },
            [
                [3.0, 3.0, 0.0, 1.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                [
                    5.3,
                    3.0,
                    2.0,
                    1.0,
                    2.0,
                    2.0,
                    1.0,
                    0.0,
                    0.0,
                ],
            ],
        ),
    ],
)
def test_inputs_get_bounds(inputs, specs, expected_bounds):
    lower, upper = inputs.get_bounds(specs=specs)
    assert np.allclose(expected_bounds[0], lower)
    assert np.allclose(expected_bounds[1], upper)


def test_inputs_get_bounds_fit():
    # at first the fix on the continuous ones is tested
    inputs = Inputs(features=[if1, if2])
    experiments = inputs.sample(100)
    experiments["if1"] += [random.uniform(-2, 2) for _ in range(100)]
    experiments["if2"] += [random.uniform(-2, 2) for _ in range(100)]
    opt_bounds = inputs.get_bounds(specs={})
    fit_bounds = inputs.get_bounds(specs={}, experiments=experiments)
    for i, key in enumerate(inputs.get_keys(ContinuousInput)):
        assert fit_bounds[0][i] < opt_bounds[0][i]
        assert fit_bounds[1][i] > opt_bounds[1][i]
        assert fit_bounds[0][i] == experiments[key].min()
        assert fit_bounds[1][i] == experiments[key].max()
    # next test the fix for the CategoricalDescriptor feature
    inputs = Inputs(
        features=[
            if1,
            if2,
            if3,
            if4,
            if5,
            if6,
        ]
    )
    experiments = inputs.sample(100)
    experiments["if4"] = [random.choice(if4.categories) for _ in range(100)]
    experiments["if6"] = [random.choice(if6.categories) for _ in range(100)]
    opt_bounds = inputs.get_bounds(
        specs={
            "if3": CategoricalEncodingEnum.ONE_HOT,
            "if4": CategoricalEncodingEnum.ONE_HOT,
            "if5": CategoricalEncodingEnum.DESCRIPTOR,
            "if6": CategoricalEncodingEnum.DESCRIPTOR,
        }
    )
    fit_bounds = inputs.get_bounds(
        {
            "if3": CategoricalEncodingEnum.ONE_HOT,
            "if4": CategoricalEncodingEnum.ONE_HOT,
            "if5": CategoricalEncodingEnum.DESCRIPTOR,
            "if6": CategoricalEncodingEnum.DESCRIPTOR,
        },
        experiments=experiments,
    )
    # check difference in descriptors
    assert opt_bounds[0][-8] == 1
    assert opt_bounds[1][-8] == 1
    assert opt_bounds[0][-7] == 2
    assert opt_bounds[1][-7] == 2
    assert fit_bounds[0][-8] == 1
    assert fit_bounds[0][-7] == 1
    assert fit_bounds[1][-8] == 5
    assert fit_bounds[1][-7] == 7
    # check difference in onehots
    assert opt_bounds[1][-1] == 0
    assert opt_bounds[1][-2] == 0
    assert fit_bounds[1][-1] == 1
    assert fit_bounds[1][-2] == 1


mixed_data = pd.DataFrame(
    columns=["of1", "of2", "of3"],
    index=range(5),
    data=np.random.uniform(size=(5, 3)),
)
mixed_data["of4"] = ["a", "a", "b", "b", "a"]


@pytest.mark.parametrize(
    "features, samples",
    [
        (
            outputs,
            pd.DataFrame(
                columns=["of1", "of2"],
                index=range(5),
                data=np.random.uniform(size=(5, 2)),
            ),
        ),
        (
            Outputs(features=[of1, of2, of3]),
            pd.DataFrame(
                columns=["of1", "of2", "of3"],
                index=range(5),
                data=np.random.uniform(size=(5, 3)),
            ),
        ),
        (
            Outputs(
                features=[
                    of1,
                    of2,
                    of3,
                    CategoricalOutput(
                        key="of4", categories=["a", "b"], objective=[1.0, 0.0]
                    ),
                ]
            ),
            mixed_data,
        ),
    ],
)
def test_outputs_call(features, samples):
    o = features(samples)
    assert o.shape == (
        len(samples),
        len(features.get_keys_by_objective(Objective))
        + len(features.get_keys(CategoricalOutput)),
    )
    assert list(o.columns) == [
        f"{key}_des"
        for key in features.get_keys_by_objective(Objective)
        + features.get_keys(CategoricalOutput)
    ]


def test_categorical_output():
    feature = CategoricalOutput(
        key="a", categories=["alpha", "beta", "gamma"], objective=[1.0, 0.0, 0.1]
    )

    assert feature.to_dict() == {"alpha": 1.0, "beta": 0.0, "gamma": 0.1}
    data = pd.Series(data=["alpha", "beta", "beta", "gamma"], name="a")
    assert_series_equal(feature(data), pd.Series(data=[1.0, 0.0, 0.0, 0.1], name="a"))



from typing import Literal, Optional, Sequence, Union

from bofire.data_models.kernels.categorical import HammondDistanceKernel
from bofire.data_models.kernels.continuous import LinearKernel, MaternKernel, RBFKernel
from bofire.data_models.kernels.kernel import Kernel
from bofire.data_models.kernels.molecular import TanimotoKernel
from bofire.data_models.priors.api import AnyPrior


class AdditiveKernel(Kernel):
    type: Literal["AdditiveKernel"] = "AdditiveKernel"
    kernels: Sequence[
        Union[
            RBFKernel,
            MaternKernel,
            LinearKernel,
            HammondDistanceKernel,
            TanimotoKernel,
            "AdditiveKernel",
            "MultiplicativeKernel",
            "ScaleKernel",
        ]
    ]
    type: Literal["AdditiveKernel"] = "AdditiveKernel"


class MultiplicativeKernel(Kernel):
    type: Literal["MultiplicativeKernel"] = "MultiplicativeKernel"
    kernels: Sequence[
        Union[
            RBFKernel,
            MaternKernel,
            LinearKernel,
            HammondDistanceKernel,
            AdditiveKernel,
            TanimotoKernel,
            "MultiplicativeKernel",
            "ScaleKernel",
        ]
    ]


class ScaleKernel(Kernel):
    type: Literal["ScaleKernel"] = "ScaleKernel"
    base_kernel: Union[
        RBFKernel,
        MaternKernel,
        LinearKernel,
        HammondDistanceKernel,
        AdditiveKernel,
        MultiplicativeKernel,
        TanimotoKernel,
        "ScaleKernel",
    ]
    outputscale_prior: Optional[AnyPrior] = None


AdditiveKernel.update_forward_refs()
MultiplicativeKernel.update_forward_refs()



import numpy as np
import pytest
import torch
from pandas.testing import assert_frame_equal
from sklearn.ensemble import RandomForestRegressor
from sklearn.exceptions import NotFittedError

import bofire.surrogates.api as surrogates
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.surrogates.api import RandomForestSurrogate
from bofire.surrogates.random_forest import _RandomForest


def test_random_forest_no_random_forest_regressor():
    with pytest.raises(ValueError):
        _RandomForest(rf=5)


def test_random_forest_not_fitted():
    with pytest.raises(NotFittedError):
        _RandomForest(rf=RandomForestRegressor())


def test_random_forest_forward():
    bench = Himmelblau()
    samples = bench.domain.inputs.sample(10)
    experiments = bench.f(samples, return_complete=True)
    rfr = RandomForestRegressor().fit(
        experiments[["x_1", "x_2"]].values, experiments.y.values.ravel()
    )
    rf = _RandomForest(rf=rfr)
    pred = rf.forward(torch.from_numpy(experiments[["x_1", "x_2"]].values))
    assert np.allclose(
        rfr.predict(experiments[["x_1", "x_2"]].values),
        pred.numpy().mean(axis=-3).ravel(),
    )
    assert np.allclose(
        rfr.predict(experiments[["x_1", "x_2"]].values),
        rf.posterior(torch.from_numpy(experiments[["x_1", "x_2"]].values))
        .mean.numpy()
        .ravel(),
    )
    assert pred.shape == torch.Size((100, 10, 1))
    # test with batches
    batch = torch.from_numpy(experiments[["x_1", "x_2"]].values).unsqueeze(0)
    pred = rf.forward(batch)
    assert pred.shape == torch.Size((1, 100, 10, 1))
    assert rf.num_outputs == 1


def test_random_forest():
    # test only continuous
    bench = Himmelblau()
    samples = bench.domain.inputs.sample(10)
    experiments = bench.f(samples, return_complete=True)
    rf = RandomForestSurrogate(inputs=bench.domain.inputs, outputs=bench.domain.outputs)
    rf = surrogates.map(rf)
    rf.fit(experiments=experiments)
    # test with categoricals
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments.loc[experiments.x_cat == "mama", "y"] *= 5.0
    experiments.loc[experiments.x_cat == "papa", "y"] /= 2.0
    experiments["valid_y"] = 1
    rf = RandomForestSurrogate(inputs=inputs, outputs=outputs)
    rf = surrogates.map(rf)
    assert rf.input_preprocessing_specs["x_cat"] == CategoricalEncodingEnum.ONE_HOT
    with pytest.raises(ValueError):
        rf.dumps()
    rf.fit(experiments=experiments)
    # test dumps and load
    preds = rf.predict(experiments)
    dump = rf.dumps()
    rf2 = RandomForestSurrogate(inputs=inputs, outputs=outputs)
    rf2 = surrogates.map(rf2)
    rf2.loads(dump)
    preds2 = rf.predict(experiments)
    assert_frame_equal(preds, preds2)



from typing import Literal, Optional

from pydantic import Field, validator
from typing_extensions import Annotated

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    NumericalInput,
)
from bofire.data_models.surrogates.surrogate import Surrogate
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class XGBoostSurrogate(Surrogate, TrainableSurrogate):
    type: Literal["XGBoostSurrogate"] = "XGBoostSurrogate"
    n_estimators: Annotated[int, Field(ge=1)] = 100
    max_depth: Annotated[int, Field(ge=0)] = 6
    max_leaves: Annotated[int, Field(ge=0)] = 0
    max_bin: Annotated[int, Field(ge=0)] = 256
    grow_policy: Literal["depthwise", "lossguide"] = "depthwise"
    learning_rate: Annotated[float, Field(gt=0.0, le=1.0)] = 0.3
    objective: Literal[
        "reg:squarederror",
        "reg:squaredlogerror",
        "reg:absoluteerror",
        "reg:pseudohubererror",
    ] = "reg:squarederror"
    booster: Literal["gbtree", "gblinear", "dart"] = "gbtree"
    n_jobs: Annotated[int, Field(gt=0)] = 1
    gamma: Annotated[float, Field(ge=0.0)] = 0.0
    min_child_weight: Annotated[float, Field(ge=0)] = 1.0
    max_delta_step: Annotated[float, Field(ge=0)] = 0.0
    subsample: Annotated[float, Field(gt=0, le=1)] = 1.0
    sampling_method: Literal["uniform", "gradient_based"] = "uniform"
    colsample_bytree: Annotated[float, Field(gt=0, le=1)] = 1.0
    colsample_bylevel: Annotated[float, Field(gt=0, le=1)] = 1.0
    colsample_bynode: Annotated[float, Field(gt=0, le=1)] = 1.0
    reg_alpha: Annotated[float, Field(ge=0)] = 0.0
    reg_lambda: Annotated[float, Field(ge=0)] = 1.0
    scale_pos_weight: Annotated[float, Field(ge=0)] = 1
    random_state: Optional[Annotated[int, Field(ge=0)]] = None
    num_parallel_tree: Annotated[int, Field(gt=0)] = 1

    @validator("input_preprocessing_specs", always=True)
    def validate_input_preprocessing_specs(cls, v, values):
        inputs = values["inputs"]
        categorical_keys = inputs.get_keys(CategoricalInput, exact=True)
        descriptor_keys = inputs.get_keys(CategoricalDescriptorInput, exact=True)
        for key in categorical_keys:
            if v.get(key, CategoricalEncodingEnum.ONE_HOT) not in [
                CategoricalEncodingEnum.ONE_HOT,
                CategoricalEncodingEnum.DUMMY,
                CategoricalEncodingEnum.ORDINAL,
            ]:
                raise ValueError(
                    "Botorch based models have to use one hot encodings for categoricals"
                )
            else:
                v[key] = CategoricalEncodingEnum.ONE_HOT
        for key in descriptor_keys:
            if v.get(key, CategoricalEncodingEnum.DESCRIPTOR) not in [
                CategoricalEncodingEnum.DESCRIPTOR,
                CategoricalEncodingEnum.ONE_HOT,
                CategoricalEncodingEnum.DUMMY,
                CategoricalEncodingEnum.ORDINAL,
            ]:
                raise ValueError(
                    "Botorch based models have to use one hot encodings or descriptor encodings for categoricals."
                )
            elif v.get(key) is None:
                v[key] = CategoricalEncodingEnum.DESCRIPTOR
        for key in inputs.get_keys(NumericalInput):
            if v.get(key) is not None:
                raise ValueError("Currently no numeric transforms are supported.")
        return v



from typing import ClassVar, Literal, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import Field, root_validator, validator

from bofire.data_models.features.feature import Output
from bofire.data_models.features.numerical import NumericalInput
from bofire.data_models.objectives.api import AnyObjective, MaximizeObjective


class ContinuousInput(NumericalInput):
    """Base class for all continuous input features.

    Attributes:
        bounds (Tuple[float, float]): A tuple that stores the lower and upper bound of the feature.
        stepsize (float, optional): Float indicating the allowed stepsize between lower and upper. Defaults to None.
    """

    type: Literal["ContinuousInput"] = "ContinuousInput"
    order_id: ClassVar[int] = 1

    bounds: Tuple[float, float]
    stepsize: Optional[float] = None

    @property
    def lower_bound(self) -> float:
        return self.bounds[0]

    @property
    def upper_bound(self) -> float:
        return self.bounds[1]

    @validator("stepsize")
    def validate_step_size(cls, v, values):
        if v is None:
            return v
        lower, upper = values["bounds"]
        if lower == upper and v is not None:
            raise ValueError(
                "Stepsize cannot be provided for a fixed continuous input."
            )
        range = upper - lower
        if np.arange(lower, upper + v, v)[-1] != upper:
            raise ValueError(
                f"Stepsize of {v} does not match the provided interval [{lower},{upper}]."
            )
        if range // v == 1:
            raise ValueError("Stepsize is too big, only one value allowed.")
        return v

    def round(self, values: pd.Series) -> pd.Series:
        """Round values to the stepsize of the feature. If no stepsize is provided return the
        provided values.

        Args:
            values (pd.Series): The values that should be rounded.

        Returns:
            pd.Series: The rounded values
        """
        if self.stepsize is None:
            return values
        self.validate_candidental(values=values)
        allowed_values = np.arange(
            self.lower_bound, self.upper_bound + self.stepsize, self.stepsize
        )
        idx = abs(values.values.reshape([3, 1]) - allowed_values).argmin(axis=1)  # type: ignore
        return pd.Series(
            data=self.lower_bound + idx * self.stepsize, index=values.index
        )

    @root_validator(pre=False, skip_on_failure=True)
    def validate_lower_upper(cls, values):
        """Validates that the lower bound is lower than the upper bound

        Args:
            values (Dict): Dictionary with attributes key, lower and upper bound

        Raises:
            ValueError: when the lower bound is higher than the upper bound

        Returns:
            Dict: The attributes as dictionary
        """
        if values["bounds"][0] > values["bounds"][1]:
            raise ValueError(
                f'lower bound must be <= upper bound, got {values["lower_bound"]} > {values["upper_bound"]}'
            )
        return values

    def validate_candidental(self, values: pd.Series) -> pd.Series:
        """Method to validate the suggested candidates

        Args:
            values (pd.Series): A dataFrame with candidates

        Raises:
            ValueError: when non numerical values are passed
            ValueError: when values are larger than the upper bound of the feature
            ValueError: when values are lower than the lower bound of the feature

        Returns:
            pd.Series: The passed dataFrame with candidates
        """

        noise = 10e-6
        values = super().validate_candidental(values)
        if (values < self.lower_bound - noise).any():
            raise ValueError(
                f"not all values of input feature `{self.key}`are larger than lower bound `{self.lower_bound}` "
            )
        if (values > self.upper_bound + noise).any():
            raise ValueError(
                f"not all values of input feature `{self.key}`are smaller than upper bound `{self.upper_bound}` "
            )
        return values

    def sample(self, n: int) -> pd.Series:
        """Draw random samples from the feature.

        Args:
            n (int): number of samples.

        Returns:
            pd.Series: drawn samples.
        """
        return pd.Series(
            name=self.key,
            data=np.random.uniform(self.lower_bound, self.upper_bound, n),
        )

    def __str__(self) -> str:
        """Method to return a string of lower and upper bound

        Returns:
            str: String of a list with lower and upper bound
        """
        return f"[{self.lower_bound},{self.upper_bound}]"


class ContinuousOutput(Output):
    """The base class for a continuous output feature

    Attributes:
        objective (objective, optional): objective of the feature indicating in which direction it should be optimzed. Defaults to `MaximizeObjective`.
    """

    type: Literal["ContinuousOutput"] = "ContinuousOutput"
    order_id: ClassVar[int] = 7
    unit: Optional[str] = None

    objective: Optional[AnyObjective] = Field(
        default_factory=lambda: MaximizeObjective(w=1.0)
    )

    def __call__(self, values: pd.Series) -> pd.Series:
        if self.objective is None:
            return pd.Series(
                data=[np.nan for _ in range(len(values))],
                index=values.index,
                name=values.name,
            )
        return self.objective(values)  # type: ignore

    def validate_experimental(self, values: pd.Series) -> pd.Series:
        try:
            values = pd.to_numeric(values, errors="raise").astype("float64")
        except ValueError:
            raise ValueError(
                f"not all values of input feature `{self.key}` are numerical"
            )
        return values

    def __str__(self) -> str:
        return "ContinuousOutputFeature"



from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel


class InputValue(BaseModel):
    """Bofire input value.

    Attributes:
        value (Union[float, str, int]): The input value.
    """

    value: str


class OutputValue(BaseModel):
    """Bofire predicted output value.

    Attributes:
        predictedValue (Value): The predicted value.
        standardDeviation (float): Standard deviation, has to be zero or larger.
        objective (float): The objective value.
    """

    predictedValue: str
    standardDeviation: Annotated[float, Field(ge=0)]
    objective: float



from typing import Literal, Type

from bofire.data_models.constraints.api import (
    Constraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
)
from bofire.data_models.features.api import Feature
from bofire.data_models.objectives.api import Objective
from bofire.data_models.strategies.strategy import Strategy


class RandomStrategy(Strategy):
    type: Literal["RandomStrategy"] = "RandomStrategy"

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        if my_type in [NChooseKConstraint, NonlinearEqualityConstraint]:
            return False
        return True

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return True

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        return True



import json
from typing import Any, Dict, Optional

import pandas as pd
from pydantic.fields import ModelField


class ValidatedDataFrame(pd.DataFrame):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        if isinstance(val, str):
            val = json.loads(val)
        if isinstance(val, dict):
            val = pd.DataFrame.from_dict(val)
        if isinstance(val, pd.DataFrame):
            val = cls(val)
        if isinstance(val, cls):
            return val
        raise TypeError("expected {cls.__name__}, pd.Dataframe, dict, or str")

    @classmethod
    def __modify_schema__(
        cls, field_schema: Dict[str, Any], field: Optional[ModelField]
    ):
        if field:
            field_schema["type"] = "object"
            field_schema["additionalProperties"] = {
                "type": "array",
                "items": {
                    "anyOf": [
                        {"type": "string"},
                        {"type": "number"},
                        {"type": "integer"},
                    ]
                },
            }

    def __eq__(self, other):
        if isinstance(other, pd.DataFrame):
            res = super().__eq__(other)
            while isinstance(res, (pd.DataFrame, pd.Series)):
                res = res.all()
            return res
        raise TypeError(f"cannot compare {self.__class__} to {other.__class__}")


class ValidatedSeries(pd.Series):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        if isinstance(val, str):
            val = json.loads(val)
        if isinstance(val, list):
            val = pd.Series(data=val)
        if isinstance(val, pd.Series):
            val = cls(val)
        if isinstance(val, cls):
            return val
        raise TypeError(f"expected {cls.__name__}, pd.Series, list, or str")

    @classmethod
    def __modify_schema__(
        cls, field_schema: Dict[str, Any], field: Optional[ModelField]
    ):
        if field:
            field_schema["type"] = "array"
            field_schema["items"] = {
                "anyOf": [
                    {"type": "string"},
                    {"type": "number"},
                    {"type": "integer"},
                ]
            }

    def __eq__(self, other):
        if isinstance(other, pd.Series):
            res = super().__eq__(other)
            while isinstance(res, pd.Series):
                res = res.all()
            return res
        raise TypeError(f"cannot compare {self.__class__} to {other.__class__}")



from typing import Literal, Optional, Union

from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class RandomForestSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["RandomForestSurrogate"] = "RandomForestSurrogate"

    # hyperparams passed down to `RandomForestRegressor`
    n_estimators: int = 100
    criterion: Literal[
        "squared_error",
        "absolute_error",
        "friedman_mse",
        "poisson",
    ] = "squared_error"
    max_depth: Optional[int] = None
    min_samples_split: Union[int, float] = 2
    min_samples_leaf: Union[int, float] = 1
    min_weight_fraction_leaf: float = 0.0
    max_features: Union[int, float, Literal["auto", "sqrt", "log2"]] = 1.0
    max_leaf_nodes: Optional[int] = None
    min_impurity_decrease: float = 0.0
    bootstrap: bool = True
    oob_score: bool = False
    random_state: Optional[int] = None
    ccp_alpha: Annotated[float, Field(ge=0)] = 0.0
    max_samples: Optional[Union[int, float]] = None



from typing import List

import torch
from botorch.acquisition.multi_objective.monte_carlo import (
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.models.gpytorch import GPyTorchModel

from bofire.data_models.strategies.api import QnehviStrategy as DataModel
from bofire.strategies.predictives.qehvi import QehviStrategy
from bofire.utils.torch_tools import get_output_constraints, tkwargs


class QnehviStrategy(QehviStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.alpha = data_model.alpha

    def _get_acqfs(self, n) -> List[qNoisyExpectedHypervolumeImprovement]:
        X_train, X_pending = self.get_acqf_input_tensors()

        # get etas and constraints
        constraints, etas = get_output_constraints(self.domain.outputs)
        if len(constraints) == 0:
            constraints, etas = None, 1e-3
        else:
            etas = torch.tensor(etas).to(**tkwargs)

        assert self.model is not None
        # if the reference point is not defined it has to be calculated from data
        acqf = qNoisyExpectedHypervolumeImprovement(
            model=self.model,
            ref_point=self.get_adjusted_refpoint(),
            X_baseline=X_train,
            # sampler=self.sampler,
            prune_baseline=True,
            objective=self._get_objective(),
            cache_root=True if isinstance(self.model, GPyTorchModel) else False,
            X_pending=X_pending,
            constraints=constraints,
            eta=etas,
            alpha=self.alpha,
        )
        acqf._default_sample_shape = torch.Size([self.num_sobol_samples])
        return [acqf]



from typing import Dict, Optional

import numpy as np
import pandas as pd
import torch
from botorch import fit_fully_bayesian_model_nuts
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.transforms.outcome import Standardize

from bofire.data_models.enum import OutputFilteringEnum
from bofire.data_models.surrogates.api import SaasSingleTaskGPSurrogate as DataModel
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.single_task_gp import get_scaler
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.utils.torch_tools import tkwargs


class SaasSingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.warmup_steps = data_model.warmup_steps
        self.num_samples = data_model.num_samples
        self.thinning = data_model.thinning
        self.scaler = data_model.scaler
        super().__init__(data_model=data_model, **kwargs)

    model: Optional[SaasFullyBayesianSingleTaskGP] = None
    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL
    training_specs: Dict = {}

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame, disable_progbar: bool = True):
        scaler = get_scaler(self.inputs, self.input_preprocessing_specs, self.scaler, X)
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)

        tX, tY = torch.from_numpy(transformed_X.values).to(**tkwargs), torch.from_numpy(
            Y.values
        ).to(**tkwargs)

        self.model = SaasFullyBayesianSingleTaskGP(
            train_X=tX,
            train_Y=tY,
            outcome_transform=Standardize(m=1),
            input_transform=scaler,
        )
        fit_fully_bayesian_model_nuts(
            self.model,
            warmup_steps=self.warmup_steps,
            num_samples=self.num_samples,
            thinning=self.thinning,
            disable_progbar=disable_progbar,
        )

    def _predict(self, transformed_X: pd.DataFrame):
        # transform to tensor
        X = torch.from_numpy(transformed_X.values).to(**tkwargs)
        with torch.no_grad():
            posterior = self.model.posterior(X=X, observation_noise=True)  # type: ignore

        preds = posterior.mixture_mean.detach().numpy()
        stds = np.sqrt(posterior.mixture_variance.detach().numpy())
        return preds, stds



from typing import Literal

from bofire.data_models.surrogates.botorch import BotorchSurrogate


class EmpiricalSurrogate(BotorchSurrogate):
    type: Literal["EmpiricalSurrogate"] = "EmpiricalSurrogate"



import numpy as np
import pytest

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.utils.reduce import (
    check_domain_for_reduction,
    check_existence_of_solution,
    reduce_domain,
    rref,
)

if1 = ContinuousInput(key="if1", bounds=(0, 0.8))

if2 = ContinuousInput(key="if2", bounds=(0, 0.8))

cif1 = CategoricalInput(key="cif1", categories=["A", "B"])

cif2 = CategoricalInput(key="cif2", categories=["A", "B"])

of1 = ContinuousOutput(key="of1")

of2 = ContinuousOutput(key="of2")


def test_check_domain_for_reduction():
    # no constraints
    domain = Domain(inputs=[if1, cif1], outputs=[of1, of2])
    assert not check_domain_for_reduction(domain)

    # no linear equality constraints
    domain = Domain(
        inputs=[if1, if2],
        outputs=[of1, of2],
        constraints=[
            LinearInequalityConstraint.from_greater_equal(
                features=["if1", "if2"], coefficients=[1.0, 1.0], rhs=0.9
            )
        ],
    )
    assert not check_domain_for_reduction(domain)

    # no continuous input
    domain = Domain(inputs=[cif1, cif2], outputs=[of1, of2])
    assert not check_domain_for_reduction(domain)

    # everything okay
    domain = Domain(
        inputs=[if1, if2],
        outputs=[of1, of2],
        constraints=[
            LinearEqualityConstraint(
                features=["if1", "if2"], coefficients=[1.0, 1.0], rhs=1.0
            )
        ],
    )
    assert check_domain_for_reduction(domain) is True


def test_check_existence_of_solution():
    # rk(A) = rk(A_aug) = 1, 3 inputs
    A_aug = np.array([[1, 0, 0, -0.5], [1, 0, 0, -0.5]])
    check_existence_of_solution(A_aug)

    # rk(A) = 1, rk(A_aug) = 2, 3 inputs
    A_aug = np.array([[1, 0, 0, 0.5], [1, 0, 0, -0.5]])
    with pytest.raises(Exception):
        check_existence_of_solution(A_aug)

    # rk(A) = rk(A_aug) = 2, 3 inputs
    A_aug = np.array([[1, 0, 0, -0.5], [0, 1, 0, -0.5]])
    check_existence_of_solution(A_aug)

    # rk(A) = rk(A_aug) = 0, 3 inputs
    A_aug = np.array([[0, 0, 0, 0], [0, 0, 0, 0]])
    check_existence_of_solution(A_aug)

    # rk(A) = rk(A_aug) = 2, 2 inputs
    A_aug = np.array([[1, 0, -0.5], [0, 1, -0.5]])
    with pytest.raises(Exception):
        check_existence_of_solution(A_aug)


def test_reduce_1_independent_linear_equality_constraints():
    # define problem: standard case
    # this is not allowed in bofire as it produces one dime constraints
    domain = Domain(
        inputs=[
            ContinuousInput(key="x1", bounds=(1, 2)),
            ContinuousInput(key="x2", bounds=(-1, 1)),
            CategoricalInput(key="x3", categories=["A", "B"]),
            CategoricalInput(key="x4", categories=["A", "B"]),
        ],
        outputs=[ContinuousOutput(key="y1")],
        constraints=[
            LinearEqualityConstraint(features=["x1", "x2"], coefficients=[1, 1], rhs=0),
            LinearEqualityConstraint(
                features=["x1", "x2"], coefficients=[-0.5, -0.5], rhs=0
            ),
            LinearInequalityConstraint.from_greater_equal(
                features=["x1", "x2"], coefficients=[-1.0, -1.0], rhs=0
            ),
        ],
    )
    _domain, transform = reduce_domain(domain)

    assert len(_domain.inputs) == 3
    assert len(_domain.constraints) == 0
    assert _domain.get_feature("x2").lower_bound == -1.0
    assert _domain.get_feature("x2").upper_bound == -1.0

    # define problem: irreducible problem
    domain = Domain(
        inputs=[
            ContinuousInput(
                key="x1",
                bounds=(1, 2),
            ),
            ContinuousInput(
                key="x2",
                bounds=(-1, 1),
            ),
            CategoricalInput(key="x3", categories=["A", "B"]),
        ],
        outputs=[ContinuousOutput(key="y1")],
    )
    assert domain == reduce_domain(domain)[0]

    # define problem: linear equality constraints can't be fulfilled inside the domain
    domain = Domain(
        inputs=[
            ContinuousInput(
                key="x1",
                bounds=(1, 2),
            ),
            ContinuousInput(
                key="x2",
                bounds=(-500, 500),
            ),
        ],
        outputs=[ContinuousOutput(key="y1")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2"], coefficients=[1.0, 0.0], rhs=0
            )
        ],
    )
    with pytest.raises(Exception):
        reduce_domain(domain)


def test_reduce_2_independent_linear_equality_constraints():
    # define problem: standard case
    domain = Domain(
        inputs=[
            ContinuousInput(
                key="x1",
                bounds=(-1, 1),
            ),
            ContinuousInput(
                key="x2",
                bounds=(-1, 1),
            ),
            ContinuousInput(
                key="x3",
                bounds=(-1, 1),
            ),
        ],
        outputs=[ContinuousOutput(key="y1")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1.0, 1.0, 1.0], rhs=1
            ),
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1.0, 2.0, 1.0], rhs=2
            ),
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[-1.0, -1.0, -1.0], rhs=-1
            ),
        ],
    )
    _domain, transform = reduce_domain(domain)

    assert len(_domain.inputs) == 1
    assert _domain.inputs[0].key == "x3"
    assert _domain.get_feature("x3").lower_bound == -1.0
    assert _domain.get_feature("x3").upper_bound == 1.0

    assert transform.equalities == [("x1", ["x3"], [-1.0, 0.0]), ("x2", [], [1.0])]


def test_reduce_3_independent_linear_equality_constraints():
    # define problem: standard case
    domain = Domain(
        inputs=[
            ContinuousInput(
                key="x1",
                bounds=(-1, 1),
            ),
            ContinuousInput(
                key="x2",
                bounds=(-1, 1),
            ),
            ContinuousInput(
                key="x3",
                bounds=(-1, 1),
            ),
        ],
        outputs=[ContinuousOutput(key="y1")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1.0, 1.0, 1.0], rhs=1
            ),
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1.0, 2.0, 1.0], rhs=2
            ),
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[0.0, 0.0, 1.0], rhs=3
            ),
        ],
    )
    with pytest.raises(Exception):
        reduce_domain(domain)


def test_doc_simple():
    domain = Domain()
    inputs = [
        ContinuousInput(
            key="x1",
            bounds=(0.1, 1),
        ),
        ContinuousInput(
            key="x2",
            bounds=(0, 0.8),
        ),
        ContinuousInput(
            key="x3",
            bounds=(0.3, 0.9),
        ),
    ]
    outputs = [ContinuousOutput(key="y")]
    domain = Domain(
        inputs=inputs,
        outputs=outputs,
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1.0, 1.0, 1.0], rhs=1
            )
        ],
    )

    _domain, transform = reduce_domain(domain)

    assert len(_domain.get_features()) == 3
    assert np.allclose(
        [_domain.get_feature("x2").lower_bound, _domain.get_feature("x2").upper_bound],
        [0.0, 0.8],
    )
    assert np.allclose(
        [_domain.get_feature("x3").lower_bound, _domain.get_feature("x3").upper_bound],
        [0.3, 0.9],
    )

    assert len(_domain.constraints) == 2

    assert all(np.array(_domain.constraints[0].features) == np.array(["x2", "x3"]))
    assert np.allclose(_domain.constraints[0].coefficients, [-1.0, -1.0])
    assert np.allclose(_domain.constraints[0].rhs, 0.0)

    assert all(np.array(_domain.constraints[1].features) == np.array(["x2", "x3"]))
    assert np.allclose(_domain.constraints[1].coefficients, [1.0, 1.0])
    assert np.allclose(_domain.constraints[1].rhs, 0.9)


def test_doc_complex():
    domain = Domain()

    inputs = [
        ContinuousInput(
            key="A1",
            bounds=(0, 0.9),
        ),
        ContinuousInput(
            key="A2",
            bounds=(0, 0.8),
        ),
        ContinuousInput(
            key="A3",
            bounds=(0, 0.9),
        ),
        ContinuousInput(
            key="A4",
            bounds=(0, 0.9),
        ),
        ContinuousInput(
            key="B1",
            bounds=(0.3, 0.9),
        ),
        ContinuousInput(
            key="B2",
            bounds=(0, 0.8),
        ),
        ContinuousInput(
            key="B3",
            bounds=(0.1, 1),
        ),
        CategoricalInput(key="Process", categories=["p1", "p2", "p3"]),
        CategoricalInput(key="Discrete", categories=["a1", "a2", "a3"]),
    ]
    constraints = [
        LinearEqualityConstraint(
            features=["A1", "A2", "A3", "A4"],
            coefficients=[1.0, 1.0, 1.0, 1.0],
            rhs=1.0,
        ),
        LinearEqualityConstraint(
            features=["B1", "B2", "B3"], coefficients=[1.0, 1.0, 1], rhs=1.0
        ),
        LinearInequalityConstraint.from_greater_equal(
            features=["A1", "A2"], coefficients=[-1.0, -2.0], rhs=-0.8
        ),
    ]
    domain = Domain(inputs=inputs, constraints=constraints)
    _domain, transform = reduce_domain(domain)

    assert len(_domain.get_features()) == 7
    assert len(_domain.get_features(ContinuousInput)) == 5

    assert np.allclose(
        [_domain.get_feature("A2").lower_bound, _domain.get_feature("A2").upper_bound],
        [0.0, 0.8],
    )
    assert np.allclose(
        [_domain.get_feature("A3").lower_bound, _domain.get_feature("A3").upper_bound],
        [0.0, 0.9],
    )
    assert np.allclose(
        [_domain.get_feature("A4").lower_bound, _domain.get_feature("A4").upper_bound],
        [0.0, 0.9],
    )
    assert np.allclose(
        [_domain.get_feature("B2").lower_bound, _domain.get_feature("B2").upper_bound],
        [0.0, 0.8],
    )
    assert np.allclose(
        [_domain.get_feature("B3").lower_bound, _domain.get_feature("B3").upper_bound],
        [0.1, 1.0],
    )
    assert all(
        np.array(_domain.constraints[0].features) == np.array(["A2", "A3", "A4"])
    )
    assert np.allclose(_domain.constraints[0].coefficients, [1.0, -1.0, -1.0])
    assert np.allclose(_domain.constraints[0].rhs, -0.2)

    assert all(
        np.array(_domain.constraints[1].features) == np.array(["A2", "A3", "A4"])
    )
    assert np.allclose(_domain.constraints[1].coefficients, [-1.0, -1.0, -1.0])
    assert np.allclose(_domain.constraints[1].rhs, -0.1)

    assert all(
        np.array(_domain.constraints[2].features) == np.array(["A2", "A3", "A4"])
    )
    assert np.allclose(_domain.constraints[2].coefficients, [1.0, 1.0, 1.0])
    assert np.allclose(_domain.constraints[2].rhs, 1.0)

    assert all(np.array(_domain.constraints[3].features) == np.array(["B2", "B3"]))
    assert np.allclose(
        _domain.constraints[3].coefficients,
        [
            -1.0,
            -1.0,
        ],
    )
    assert np.allclose(_domain.constraints[3].rhs, -0.1)

    assert all(np.array(_domain.constraints[4].features) == np.array(["B2", "B3"]))
    assert np.allclose(
        _domain.constraints[4].coefficients,
        [
            1.0,
            1.0,
        ],
    )
    assert np.allclose(_domain.constraints[4].rhs, 0.7)


# def test_AffineTransform_augment_data():
#    domain = Domain()
#    domain.add_feature(
#        ContinuousInputFeature(key="x1", bounds=(0.1, 1),)
#    )
#    domain.add_feature(
#        ContinuousInputFeature(key="x2", bounds=(0, 0.8),)
#    )
#    domain.add_feature(
#        ContinuousInputFeature(key="x3", bounds=(0.3, 0.9),)
#    )
#    domain.add_feature(ContinuousOutputFeature(key="y"))
#    domain.add_constraint(
#        LinearEqualityConstraint(
#            features=["x1", "x2", "x3"], coefficients=[1.0, 1.0, 1.0], rhs=1
#        )
#    )
#
#    _domain, transform = reduce_domain(domain)
#
#    rec = RandomStrategy.from_domain(domain)
#    data, _ = rec.ask(4)
#    data.drop(columns=["y_pred", "y_sd", "y_des"], inplace=True)
#    data["y"] = [random.random() for _ in range(4)]
#    data["valid_y"] = 1
#    domain.validate_experiments(data)
#
#    data_rec = transform.augment_data(data[_domain.get_feature_keys()])
#    assert np.allclose(
#        data[domain.get_feature_keys()].values,
#        data_rec[domain.get_feature_keys()].values,
#    )
#    domain.validate_experiments(data_rec)


# def test_AffineTransform_drop_data():
#    # define problem and transform
#    domain = Domain()
#    domain.add_feature(
#        ContinuousInputFeature(key="x1", bounds=(0.1, 1),)
#    )
#    domain.add_feature(
#        ContinuousInputFeature(key="x2", bounds=(0, 0.8),)
#    )
#    domain.add_feature(
#        ContinuousInputFeature(key="x3", bounds=(0.3, 0.9),)
#    )
#    domain.add_feature(ContinuousOutputFeature(key="y"))
#    domain.add_constraint(
#        LinearEqualityConstraint(
#            features=["x1", "x2", "x3"], coefficients=[1.0, 1.0, 1.0], rhs=1
#        )
#    )
#
#    _domain, transform = reduce_domain(domain)
#
#    rec = RandomStrategy.from_domain(domain)
#    data, _ = rec.ask(4)
#    data.drop(columns=["y_pred", "y_sd", "y_des"], inplace=True)
#    data["y"] = [random.random() for _ in range(4)]
#    data["valid_y"] = 1
#    domain.validate_experiments(data)
#
#    _data = transform.drop_data(data)
#
#    with pytest.raises(ValueError):
#        domain.validate_experiments(_data)
#
#    _domain.validate_experiments(_data)
#
#    assert np.allclose(
#        data[_domain.get_feature_keys()].values,
#        _data[_domain.get_feature_keys()].values,
#    )


def test_reduce_large_problem():
    domain = Domain(
        inputs=[
            ContinuousInput(
                key="x1",
                bounds=(-1, 1),
            ),
            ContinuousInput(
                key="x2",
                bounds=(-5000, 1),
            ),
            ContinuousInput(
                key="x3",
                bounds=(-5000, 5000),
            ),
            ContinuousInput(
                key="x4",
                bounds=(-1, 1),
            ),
        ],
        outputs=[ContinuousOutput(key="y1")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x4"], coefficients=[1.0, -1.0, 1.0], rhs=-1.0
            ),
            LinearEqualityConstraint(
                features=["x2", "x3"], coefficients=[2, 1], rhs=2.0
            ),
            LinearEqualityConstraint(
                features=["x1", "x2", "x3", "x4"],
                coefficients=[1.0, 1.0, 1.0, 1.0],
                rhs=1.0,
            ),
            LinearInequalityConstraint.from_greater_equal(
                features=["x1", "x2"], coefficients=[-1.0, -1.0], rhs=-1.0
            ),
            LinearInequalityConstraint.from_greater_equal(
                features=["x1", "x2", "x4"], coefficients=[-1.0, 1.0, -1.0], rhs=0.0
            ),
        ],
    )
    _domain, transform = reduce_domain(domain)

    assert transform.equalities == [
        ("x1", ["x3", "x4"], [-0.5, -1.0, 0.0]),
        ("x2", ["x3"], [-0.5, 1.0]),
    ]
    assert len(_domain.constraints) == 3
    assert all(np.array(_domain.constraints[0].features) == np.array(["x3", "x4"]))
    assert np.allclose(_domain.constraints[0].coefficients, [-1.0, -1.0])
    assert np.allclose(_domain.constraints[0].rhs, 0.0)

    assert all(np.array(_domain.constraints[1].features) == np.array(["x3", "x4"]))
    assert np.allclose(_domain.constraints[1].coefficients, [-0.5, -1.0])
    assert np.allclose(_domain.constraints[1].rhs, 1.0)

    assert all(np.array(_domain.constraints[2].features) == np.array(["x3", "x4"]))
    assert np.allclose(_domain.constraints[2].coefficients, [0.5, 1.0])
    assert np.allclose(_domain.constraints[2].rhs, 1.0)

    assert np.allclose(
        [_domain.get_feature("x3").lower_bound, _domain.get_feature("x3").upper_bound],
        [-0.0, 5000.0],
    )
    assert np.allclose(
        [_domain.get_feature("x4").lower_bound, _domain.get_feature("x4").upper_bound],
        [-1.0, 1.0],
    )


def test_rref():
    A = np.vstack(([[np.pi, 1e10, np.log(10), 7]], [[np.pi, 1e10, np.log(10), 7]]))
    A_rref, pivots = rref(A)
    B_rref = np.array(
        [
            [1.0, 3183098861.837907, 0.7329355988794278, 2.228169203286535],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([0]))

    A = np.array(
        [
            [np.pi, 2, 1, 1, 1],
            [1e10, np.exp(0), 2, 2, 2],
            [np.log(10), -5.2, 3, 3, 3],
            [7, -3.5 * 1e-4, 4, 4, 4],
        ]
    )
    A_rref, pivots = rref(A)
    B_rref = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 1.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([0, 1, 2]))

    A = np.ones(shape=(100, 100))
    A_rref, pivots = rref(A)
    B_rref = np.zeros(shape=(100, 100))
    B_rref[0, :] = 1
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([0]))

    A = np.zeros(shape=(10, 20))
    A_rref, pivots = rref(A)
    B_rref = np.zeros(shape=(10, 20))
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([]))

    A = np.array([[0, 1, 2, 2, 3, 4], [0, 5, 10, 6, 7, 8], [0, 9, 18, 10, 11, 12]])
    A_rref, pivots = rref(A)
    B_rref = np.array([[0, 1, 2, 0, -1, -2], [0, 0, 0, 1, 2, 3], [0, 0, 0, 0, 0, 0]])
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([1, 3]))

    A = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
    A_rref, pivots = rref(A)
    B_rref = np.array([[1, 0, -1, -2], [0, 1, 2, 3], [0, 0, 0, 0]])
    assert np.all(np.round(A_rref, 8) == np.round(B_rref, 8))
    assert all(np.array(pivots) == np.array([0, 1]))



from bofire.data_models import unions
from bofire.data_models.acquisition_functions.api import (
    AcquisitionFunction,
    AnyAcquisitionFunction,
)
from bofire.data_models.constraints.api import AnyConstraint, Constraint
from bofire.data_models.domain.api import Domain, Features, Inputs, Outputs
from bofire.data_models.features.api import (
    AnyFeature,
    AnyInput,
    AnyOutput,
    Feature,
    Input,
    Output,
)

try:
    # in case of the minimal installation these import are not available
    from bofire.data_models.kernels.api import AnyKernel, Kernel
    from bofire.data_models.molfeatures.api import (  # noqa: F401
        AnyMolFeatures,
        MolFeatures,
    )
    from bofire.data_models.objectives.api import AnyObjective, Objective
    from bofire.data_models.outlier_detection.api import (
        AnyOutlierDetection,
        OutlierDetection,
    )
    from bofire.data_models.priors.api import AnyPrior, Prior
    from bofire.data_models.strategies.api import (
        AnyCondition,
        AnyPredictive,
        AnySampler,
        AnyStrategy,
        PredictiveStrategy,
        SamplerStrategy,
        Strategy,
    )
    from bofire.data_models.surrogates.api import (
        AnyBotorchSurrogate,
        AnySurrogate,
        BotorchSurrogate,
        Surrogate,
    )

    data_model_list = [
        AnyAcquisitionFunction,
        AnyConstraint,
        AnyFeature,
        AnyKernel,
        AnySurrogate,
        AnyOutlierDetection,
        AnyObjective,
        AnyPrior,
        AnyStrategy,
        AnyMolFeatures,
        Domain,
    ]
except ImportError:
    data_model_list = [
        AnyAcquisitionFunction,
        AnyConstraint,
        AnyFeature,
        Domain,
    ]

AnyThing = [model for models in data_model_list for model in unions.to_list(models)]

AnyThing = [model for models in data_model_list for model in unions.to_list(models)]



import pytest

from bofire.benchmarks.multi import C2DTLZ2, DTLZ2, ZDT1, CrossCoupling, SnarBenchmark


@pytest.mark.parametrize(
    "cls_benchmark, return_complete, kwargs",
    [
        (DTLZ2, True, {"dim": 5}),
        (SnarBenchmark, True, {}),
        (ZDT1, True, {}),
        (
            CrossCoupling,
            True,
            {},
        ),
        (DTLZ2, False, {"dim": 5}),
        (SnarBenchmark, False, {}),
        (ZDT1, False, {}),
        (
            CrossCoupling,
            False,
            {},
        ),
        (C2DTLZ2, True, {"dim": 4}),
        (C2DTLZ2, False, {"dim": 4}),
    ],
)
def test_multi_objective_benchmarks(cls_benchmark, return_complete, kwargs):
    """Test function for multi objective benchmark functions.

    Args:
        cls_benchmark (Benchmark function class): Benchmark function that is supposed to be tested.
        kwargs ({"dim": , "k":}): Optinal arguments for benchmark functions that require additional arguments. DTLZ2 requires "dim" and "k".
    """
    benchmark_function = cls_benchmark(**kwargs)
    benchmark_function_name = benchmark_function.__class__.__name__

    # Check for correct output dimensions
    n_samples = 1000
    X_samples = benchmark_function.domain.inputs.sample(n=n_samples)
    Y = benchmark_function.f(X_samples, return_complete=return_complete)

    # Define expected number of output variables
    expected_output_variables = len(benchmark_function.domain.outputs) * 2
    # Check, whether expected number of output variables match the actual number
    if return_complete:
        assert Y.shape == (
            n_samples,
            len(benchmark_function.domain.experiment_column_names),
        ), (
            "The shape of the output dataframe of "
            + benchmark_function_name
            + " does not match the expected shape of "
            + "("
            + str(n_samples)
            + ","
            + str(expected_output_variables)
            + ")"
        )
    else:
        assert Y.shape == (n_samples, expected_output_variables), (
            "The shape of the output dataframe of "
            + benchmark_function_name
            + " does not match the expected shape of "
            + "("
            + str(n_samples)
            + ","
            + str(expected_output_variables)
            + ")"
        )

    # Check for correct optima/pareto front for multiobjective functions that return a vector
    # What are the corresponding X vectors to the multiobjective optima space of the input space? -> Not implemented yet in the multiobjective function



from typing import Annotated, Literal

from pydantic import Field, PositiveFloat

from bofire.data_models.base import BaseModel


class AcquisitionFunction(BaseModel):
    type: str


class SingleObjectiveAcquisitionFunction(AcquisitionFunction):
    type: str


class MultiObjectiveAcquisitionFunction(AcquisitionFunction):
    type: str


class qNEI(SingleObjectiveAcquisitionFunction):
    type: Literal["qNEI"] = "qNEI"
    prune_baseline: bool = True


class qLogNEI(SingleObjectiveAcquisitionFunction):
    type: Literal["qLogNEI"] = "qLogNEI"
    prune_baseline: bool = True


class qEI(SingleObjectiveAcquisitionFunction):
    type: Literal["qEI"] = "qEI"


class qLogEI(SingleObjectiveAcquisitionFunction):
    type: Literal["qLogEI"] = "qLogEI"


class qSR(SingleObjectiveAcquisitionFunction):
    type: Literal["qSR"] = "qSR"


class qUCB(SingleObjectiveAcquisitionFunction):
    type: Literal["qUCB"] = "qUCB"
    beta: Annotated[float, Field(ge=0)] = 0.2


class qPI(SingleObjectiveAcquisitionFunction):
    type: Literal["qPI"] = "qPI"
    tau: PositiveFloat = 1e-3


class qEHVI(MultiObjectiveAcquisitionFunction):
    type: Literal["qEHVI"] = "qEHVI"
    alpha: Annotated[float, Field(ge=0)] = 0.0


class qLogEHVI(MultiObjectiveAcquisitionFunction):
    type: Literal["qLogEHVI"] = "qLogEHVI"
    alpha: Annotated[float, Field(ge=0)] = 0.0


class qNEHVI(MultiObjectiveAcquisitionFunction):
    type: Literal["qNEHVI"] = "qNEHVI"
    alpha: Annotated[float, Field(ge=0)] = 0.0
    prune_baseline: bool = True


class qLogNEHVI(MultiObjectiveAcquisitionFunction):
    type: Literal["qLogNEHVI"] = "qLogNEHVI"
    alpha: Annotated[float, Field(ge=0)] = 0.0
    prune_baseline: bool = True



import math
from itertools import chain

import pytest
import torch
from botorch.acquisition import (
    qExpectedImprovement,
    qLogExpectedImprovement,
    qLogNoisyExpectedImprovement,
    qNoisyExpectedImprovement,
    qProbabilityOfImprovement,
    qSimpleRegret,
    qUpperConfidenceBound,
)
from botorch.acquisition.objective import ConstrainedMCObjective, GenericMCObjective

import bofire.data_models.strategies.api as data_models
import tests.bofire.data_models.specs.api as specs
from bofire.benchmarks.multi import DTLZ2
from bofire.benchmarks.single import Himmelblau, _CategoricalDiscreteHimmelblau
from bofire.data_models.acquisition_functions.api import (
    SingleObjectiveAcquisitionFunction,
    qEI,
    qLogEI,
    qLogNEI,
    qNEI,
    qPI,
    qSR,
    qUCB,
)
from bofire.data_models.domain.api import Domain, Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.objectives.api import (
    MaximizeObjective,
    MaximizeSigmoidObjective,
)
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.api import CustomSoboStrategy, PolytopeSampler, SoboStrategy
from tests.bofire.strategies.test_base import domains

# from tests.bofire.strategies.botorch.test_model_spec import VALID_MODEL_SPEC_LIST

VALID_BOTORCH_SOBO_STRATEGY_SPEC = {
    "domain": domains[1],
    "acquisition_function": specs.acquisition_functions.valid(
        SingleObjectiveAcquisitionFunction, exact=False
    ).obj(),
    # "num_sobol_samples": 1024,
    # "num_restarts": 8,
    # "num_raw_samples": 1024,
    "descriptor_method": "EXHAUSTIVE",
    "categorical_method": "EXHAUSTIVE",
}

BOTORCH_SOBO_STRATEGY_SPECS = {
    "valids": [
        VALID_BOTORCH_SOBO_STRATEGY_SPEC,
        {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "seed": 1},
        # {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "surrogate_specs": VALID_MODEL_SPEC_LIST},
    ],
    "invalids": [
        {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "acquisition_function": None},
        {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "descriptor_method": None},
        {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "categorical_method": None},
        {**VALID_BOTORCH_SOBO_STRATEGY_SPEC, "seed": -1},
    ],
}

VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC = {
    "domain": domains[2],
    "acquisition_function": specs.acquisition_functions.valid(
        SingleObjectiveAcquisitionFunction, exact=False
    ).obj(),
    "descriptor_method": "EXHAUSTIVE",
    "categorical_method": "EXHAUSTIVE",
}

BOTORCH_ADDITIVE_AND_MULTIPLICATIVE_SOBO_STRATEGY_SPECS = {
    "valids": [
        VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC,
        {**VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC, "seed": 1},
    ],
    "invalids": [
        {
            **VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC,
            "acquisition_function": None,
        },
        {
            **VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC,
            "descriptor_method": None,
        },
        {
            **VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC,
            "categorical_method": None,
        },
        {**VALID_ADDITIVE_AND_MULTIPLICATIVE_BOTORCH_SOBO_STRATEGY_SPEC, "seed": -1},
    ],
}


@pytest.mark.parametrize(
    "domain, acqf",
    [(domains[0], VALID_BOTORCH_SOBO_STRATEGY_SPEC["acquisition_function"])],
)
def test_SOBO_not_fitted(domain, acqf):
    data_model = data_models.SoboStrategy(domain=domain, acquisition_function=acqf)
    strategy = SoboStrategy(data_model=data_model)

    msg = "Model not trained."
    with pytest.raises(AssertionError, match=msg):
        strategy._get_acqfs(2)


@pytest.mark.parametrize(
    "acqf, expected, num_test_candidates",
    [
        (acqf_inp[0], acqf_inp[1], num_test_candidates)
        for acqf_inp in [
            (qEI(), qExpectedImprovement),
            (qNEI(), qNoisyExpectedImprovement),
            (qPI(), qProbabilityOfImprovement),
            (qUCB(), qUpperConfidenceBound),
            (qSR(), qSimpleRegret),
            (qLogEI(), qLogExpectedImprovement),
            (qLogNEI(), qLogNoisyExpectedImprovement),
        ]
        for num_test_candidates in range(1, 3)
    ],
)
def test_SOBO_get_acqf(acqf, expected, num_test_candidates):
    # generate data
    benchmark = Himmelblau()

    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )

    experiments = benchmark.f(random_strategy.ask(20), return_complete=True)

    data_model = data_models.SoboStrategy(
        domain=benchmark.domain, acquisition_function=acqf
    )
    strategy = SoboStrategy(data_model=data_model)

    strategy.tell(experiments)

    acqfs = strategy._get_acqfs(2)
    assert len(acqfs) == 1

    assert isinstance(acqfs[0], expected)


def test_SOBO_calc_acquisition():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(10), return_complete=True)
    samples = benchmark.domain.inputs.sample(2)
    data_model = data_models.SoboStrategy(
        domain=benchmark.domain, acquisition_function=qEI()
    )
    strategy = SoboStrategy(data_model=data_model)
    strategy.tell(experiments=experiments)
    vals = strategy.calc_acquisition(samples)
    assert len(vals) == 2
    vals = strategy.calc_acquisition(samples, combined=True)
    assert len(vals) == 1


def test_SOBO_init_qUCB():
    beta = 0.5
    acqf = qUCB(beta=beta)

    # generate data
    benchmark = Himmelblau()
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=20), return_complete=True)

    data_model = data_models.SoboStrategy(
        domain=benchmark.domain, acquisition_function=acqf
    )
    strategy = SoboStrategy(data_model=data_model)
    strategy.tell(experiments)

    acqf = strategy._get_acqfs(2)[0]
    assert acqf.beta_prime == math.sqrt(beta * math.pi / 2)


@pytest.mark.parametrize(
    "acqf, num_experiments, num_candidates",
    [
        (acqf.obj(), num_experiments, num_candidates)
        for acqf in specs.acquisition_functions.valids
        for num_experiments in range(8, 10)
        for num_candidates in range(1, 3)
    ],
)
@pytest.mark.slow
def test_get_acqf_input(acqf, num_experiments, num_candidates):
    # generate data
    benchmark = Himmelblau()
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(
        random_strategy._ask(n=num_experiments), return_complete=True
    )

    data_model = data_models.SoboStrategy(
        domain=benchmark.domain, acquisition_function=acqf
    )
    strategy = SoboStrategy(data_model=data_model)

    strategy.tell(experiments)
    strategy.ask(candidate_count=num_candidates, add_pending=True)

    X_train, X_pending = strategy.get_acqf_input_tensors()

    _, names = strategy.domain.inputs._get_transform_info(
        specs=strategy.surrogate_specs.input_preprocessing_specs
    )

    assert torch.is_tensor(X_train)
    assert torch.is_tensor(X_pending)
    assert X_train.shape == (
        num_experiments,
        len(set(chain(*names.values()))),
    )
    assert X_pending.shape == (  # type: ignore
        num_candidates,
        len(set(chain(*names.values()))),
    )


def test_custom_get_objective():
    def f(samples, callables, weights, X):
        outputs_list = []
        for c, w in zip(callables, weights):
            outputs_list.append(c(samples, None) ** w)
        samples = torch.stack(outputs_list, dim=-1)

        return (samples[..., 0] + samples[..., 1]) * (samples[..., 0] * samples[..., 1])

    benchmark = DTLZ2(3)
    data_model = data_models.CustomSoboStrategy(
        domain=benchmark.domain, acquisition_function=qNEI()
    )
    strategy = CustomSoboStrategy(data_model=data_model)
    strategy.f = f
    generic_objective, _, _ = strategy._get_objective_and_constraints()
    assert isinstance(generic_objective, GenericMCObjective)


def test_custom_get_objective_invalid():
    benchmark = DTLZ2(3)
    data_model = data_models.CustomSoboStrategy(
        domain=benchmark.domain, acquisition_function=qNEI()
    )
    strategy = CustomSoboStrategy(data_model=data_model)

    with pytest.raises(ValueError):
        strategy._get_objective_and_constraints()


def test_custom_dumps_loads():
    def f(samples, callables, weights, X):
        outputs_list = []
        for c, w in zip(callables, weights):
            outputs_list.append(c(samples, None) ** w)
        samples = torch.stack(outputs_list, dim=-1)

        return (samples[..., 0] + samples[..., 1]) * (samples[..., 0] * samples[..., 1])

    benchmark = DTLZ2(3)
    data_model1 = data_models.CustomSoboStrategy(
        domain=benchmark.domain,
        acquisition_function=qNEI(),
        use_output_constraints=False,
    )
    strategy1 = CustomSoboStrategy(data_model=data_model1)
    strategy1.f = f
    f_str = strategy1.dumps()

    data_model2 = data_models.CustomSoboStrategy(
        domain=benchmark.domain,
        acquisition_function=qNEI(),
        use_output_constraints=False,
        dump=f_str,
    )
    strategy2 = CustomSoboStrategy(data_model=data_model2)

    data_model3 = data_models.CustomSoboStrategy(
        domain=benchmark.domain, acquisition_function=qNEI()
    )
    strategy3 = CustomSoboStrategy(data_model=data_model3)
    strategy3.loads(f_str)

    assert isinstance(strategy2.f, type(f))
    assert isinstance(strategy3.f, type(f))

    samples = torch.rand(30, 2, requires_grad=True) * 5
    objective1, _, _ = strategy1._get_objective_and_constraints()
    output1 = objective1.forward(samples)
    objective2, _, _ = strategy2._get_objective_and_constraints()
    output2 = objective2.forward(samples)
    objective3, _, _ = strategy3._get_objective_and_constraints()
    output3 = objective3.forward(samples)

    torch.testing.assert_close(output1, output2)
    torch.testing.assert_close(output1, output3)


def test_custom_dumps_invalid():
    benchmark = DTLZ2(3)
    data_model = data_models.CustomSoboStrategy(
        domain=benchmark.domain, acquisition_function=qNEI()
    )
    strategy = CustomSoboStrategy(data_model=data_model)
    with pytest.raises(ValueError):
        strategy.dumps()


@pytest.mark.parametrize("candidate_count", [1, 2])
def test_sobo_fully_combinatorical(candidate_count):
    benchmark = _CategoricalDiscreteHimmelblau()

    strategy_data = data_models.SoboStrategy(domain=benchmark.domain)
    strategy = SoboStrategy(data_model=strategy_data)

    experiments = benchmark.f(benchmark.domain.inputs.sample(10), return_complete=True)

    strategy.tell(experiments=experiments)
    strategy.ask(candidate_count=candidate_count)


@pytest.mark.parametrize(
    "outputs, expected_objective",
    [
        (
            Outputs(
                features=[ContinuousOutput(key="alpha", objective=MaximizeObjective())]
            ),
            GenericMCObjective,
        ),
        (
            Outputs(
                features=[
                    ContinuousOutput(
                        key="alpha",
                        objective=MaximizeSigmoidObjective(steepness=1, tp=1),
                    )
                ]
            ),
            GenericMCObjective,
        ),
    ],
)
def test_sobo_get_obective(outputs, expected_objective):
    strategy_data = data_models.SoboStrategy(
        domain=Domain(
            inputs=Inputs(features=[ContinuousInput(key="a", bounds=(0, 1))]),
            outputs=outputs,
        )
    )
    strategy = SoboStrategy(data_model=strategy_data)
    obj, _, _ = strategy._get_objective_and_constraints()
    assert isinstance(obj, expected_objective)


def test_sobo_get_constrained_objective():
    benchmark = DTLZ2(dim=6)
    experiments = benchmark.f(benchmark.domain.inputs.sample(5), return_complete=True)
    domain = benchmark.domain
    domain.outputs.get_by_key("f_1").objective = MaximizeSigmoidObjective(
        tp=1.5, steepness=2.0
    )
    strategy_data = data_models.SoboStrategy(domain=domain, acquisition_function=qUCB())
    strategy = SoboStrategy(data_model=strategy_data)
    strategy.tell(experiments=experiments)
    obj, _, _ = strategy._get_objective_and_constraints()
    assert isinstance(obj, ConstrainedMCObjective)


def test_sobo_get_constrained_objective2():
    benchmark = DTLZ2(dim=6)
    experiments = benchmark.f(benchmark.domain.inputs.sample(5), return_complete=True)
    domain = benchmark.domain
    domain.outputs.get_by_key("f_1").objective = MaximizeSigmoidObjective(
        tp=1.5, steepness=2.0
    )
    strategy_data = data_models.SoboStrategy(domain=domain, acquisition_function=qEI())
    strategy = SoboStrategy(data_model=strategy_data)
    strategy.tell(experiments=experiments)
    obj, _, _ = strategy._get_objective_and_constraints()
    assert isinstance(obj, GenericMCObjective)


def test_sobo_hyperoptimize():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(3), return_complete=True)
    strategy_data = data_models.SoboStrategy(
        domain=benchmark.domain, acquisition_function=qEI(), frequency_hyperopt=1
    )
    strategy_data.surrogate_specs.surrogates[0].hyperconfig = None
    strategy = SoboStrategy(data_model=strategy_data)
    with pytest.warns(
        match="No hyperopt is possible as no hyperopt config is available. Returning initial config."
    ):
        strategy.tell(experiments=experiments)
